/*
 * Copyright (c) 2024 Gnattu OC <gnattuoc@me.com>
 *
 * This file is part of FFmpeg.
 *
 * FFmpeg is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * FFmpeg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with FFmpeg; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

#include "vf_tonemapx_intrin_avx.h"

#ifdef ENABLE_TONEMAPX_AVX_INTRINSICS
#    include <immintrin.h>
#endif // ENABLE_TONEMAPX_AVX_INTRINSICS

#ifdef ENABLE_TONEMAPX_AVX_INTRINSICS
X86_64_V3 static inline __m256i av_clip_int16_avx(__m256i a)
{
__m256i add_result = _mm256_add_epi32(a, _mm256_set1_epi32(0x8000U));
__m256i mask = _mm256_set1_epi32(~0xFFFF);
__m256i condition = _mm256_and_si256(add_result, mask);
__m256i cmp = _mm256_cmpeq_epi32(condition, _mm256_setzero_si256());

__m256i shifted = _mm256_srai_epi32(a, 31);
__m256i xor_result = _mm256_xor_si256(shifted, _mm256_set1_epi32(0x7FFF));

return _mm256_or_si256(_mm256_and_si256(cmp, a), _mm256_andnot_si256(cmp, xor_result));
}

X86_64_V3 static inline void tonemap_int32x8_avx(__m256i r_in, __m256i g_in, __m256i b_in,
                                                 int16_t *r_out, int16_t *g_out, int16_t *b_out,
                                                 float *lin_lut, float *tonemap_lut, uint16_t *delin_lut,
                                                 const AVLumaCoefficients *coeffs,
                                                 const AVLumaCoefficients *ocoeffs, double desat,
                                                 double (*rgb2rgb)[3][3],
                                                 int rgb2rgb_passthrough)
{
    __m256i sig8;
    __m256 mapvalx8, r_linx8, g_linx8, b_linx8;
    __m256 offset = _mm256_set1_ps(0.5f);
    __m256i zerox8 = _mm256_setzero_si256();
    __m256i input_lut_offset = _mm256_set1_epi32(2048);
    __m256i upper_bound = _mm256_set1_epi32(32767);
    __m256 intermediate_upper_bound = _mm256_set1_ps(32767.0f);
    __m256i r, g, b, rx8, gx8, bx8;

    float mapval8[8], r_lin8[8], g_lin8[8], b_lin8[8];

    sig8 = _mm256_max_epi32(r_in, _mm256_max_epi32(g_in, b_in));
    sig8 = _mm256_add_epi32(sig8, input_lut_offset);
    sig8 = _mm256_min_epi32(sig8, upper_bound);
    sig8 = _mm256_max_epi32(sig8, zerox8);

    r = _mm256_add_epi32(r_in, input_lut_offset);
    r = _mm256_min_epi32(r, upper_bound);
    r = _mm256_max_epi32(r, zerox8);
    g = _mm256_add_epi32(g_in, input_lut_offset);
    g = _mm256_min_epi32(g, upper_bound);
    g = _mm256_max_epi32(g, zerox8);
    b = _mm256_add_epi32(b_in, input_lut_offset);
    b = _mm256_min_epi32(b, upper_bound);
    b = _mm256_max_epi32(b, zerox8);

#define LOAD_LUT(i) mapval8[i] = tonemap_lut[_mm256_extract_epi32(sig8, i)]; \
r_lin8[i] = lin_lut[_mm256_extract_epi32(r, i)];                             \
g_lin8[i] = lin_lut[_mm256_extract_epi32(g, i)];                             \
b_lin8[i] = lin_lut[_mm256_extract_epi32(b, i)];

    LOAD_LUT(0)
    LOAD_LUT(1)
    LOAD_LUT(2)
    LOAD_LUT(3)
    LOAD_LUT(4)
    LOAD_LUT(5)
    LOAD_LUT(6)
    LOAD_LUT(7)

#undef LOAD_LUT

    mapvalx8 = _mm256_loadu_ps(mapval8);
    r_linx8 = _mm256_loadu_ps(r_lin8);
    g_linx8 = _mm256_loadu_ps(g_lin8);
    b_linx8 = _mm256_loadu_ps(b_lin8);

    if (!rgb2rgb_passthrough) {
        r_linx8 = _mm256_mul_ps(r_linx8, _mm256_set1_ps((float)(*rgb2rgb)[0][0]));
        r_linx8 = _mm256_fmadd_ps(g_linx8, _mm256_set1_ps((float)(*rgb2rgb)[0][1]), r_linx8);
        r_linx8 = _mm256_fmadd_ps(b_linx8, _mm256_set1_ps((float)(*rgb2rgb)[0][2]), r_linx8);

        g_linx8 = _mm256_mul_ps(g_linx8, _mm256_set1_ps((float)(*rgb2rgb)[1][1]));
        g_linx8 = _mm256_fmadd_ps(r_linx8, _mm256_set1_ps((float)(*rgb2rgb)[1][0]), g_linx8);
        g_linx8 = _mm256_fmadd_ps(b_linx8, _mm256_set1_ps((float)(*rgb2rgb)[1][2]), g_linx8);

        b_linx8 = _mm256_mul_ps(b_linx8, _mm256_set1_ps((float)(*rgb2rgb)[2][2]));
        b_linx8 = _mm256_fmadd_ps(r_linx8, _mm256_set1_ps((float)(*rgb2rgb)[2][0]), b_linx8);
        b_linx8 = _mm256_fmadd_ps(g_linx8, _mm256_set1_ps((float)(*rgb2rgb)[2][1]), b_linx8);
    }

    if (desat > 0) {
        __m256 eps_x8 = _mm256_set1_ps(FLOAT_EPS);
        __m256 desat8 = _mm256_set1_ps((float)desat);
        __m256 luma8 = _mm256_set1_ps(0);
        __m256 overbright8;

        luma8 = _mm256_fmadd_ps(r_linx8, _mm256_set1_ps((float)av_q2d(coeffs->cr)), luma8);
        luma8 = _mm256_fmadd_ps(g_linx8, _mm256_set1_ps((float)av_q2d(coeffs->cg)), luma8);
        luma8 = _mm256_fmadd_ps(b_linx8, _mm256_set1_ps((float)av_q2d(coeffs->cb)), luma8);
        overbright8 = _mm256_div_ps(_mm256_max_ps(_mm256_sub_ps(luma8, desat8), eps_x8), _mm256_max_ps(luma8, eps_x8));
        r_linx8 = _mm256_fnmadd_ps(r_linx8, overbright8, r_linx8);
        r_linx8 = _mm256_fmadd_ps(luma8, overbright8, r_linx8);
        g_linx8 = _mm256_fnmadd_ps(g_linx8, overbright8, g_linx8);
        g_linx8 = _mm256_fmadd_ps(luma8, overbright8, g_linx8);
        b_linx8 = _mm256_fnmadd_ps(b_linx8, overbright8, b_linx8);
        b_linx8 = _mm256_fmadd_ps(luma8, overbright8, b_linx8);
    }

    r_linx8 = _mm256_mul_ps(r_linx8, mapvalx8);
    g_linx8 = _mm256_mul_ps(g_linx8, mapvalx8);
    b_linx8 = _mm256_mul_ps(b_linx8, mapvalx8);

    r_linx8 = _mm256_fmadd_ps(r_linx8, intermediate_upper_bound, offset);
    g_linx8 = _mm256_fmadd_ps(g_linx8, intermediate_upper_bound, offset);
    b_linx8 = _mm256_fmadd_ps(b_linx8, intermediate_upper_bound, offset);

    rx8 = _mm256_cvttps_epi32(r_linx8);
    rx8 = _mm256_min_epi32(rx8, upper_bound);
    rx8 = _mm256_max_epi32(rx8, zerox8);

    gx8 = _mm256_cvttps_epi32(g_linx8);
    gx8 = _mm256_min_epi32(gx8, upper_bound);
    gx8 = _mm256_max_epi32(gx8, zerox8);

    bx8 = _mm256_cvttps_epi32(b_linx8);
    bx8 = _mm256_min_epi32(bx8, upper_bound);
    bx8 = _mm256_max_epi32(bx8, zerox8);

#define SAVE_COLOR(i) r_out[i] = delin_lut[_mm256_extract_epi32(rx8, i)]; \
g_out[i] = delin_lut[_mm256_extract_epi32(gx8, i)];                       \
b_out[i] = delin_lut[_mm256_extract_epi32(bx8, i)];

    SAVE_COLOR(0)
    SAVE_COLOR(1)
    SAVE_COLOR(2)
    SAVE_COLOR(3)
    SAVE_COLOR(4)
    SAVE_COLOR(5)
    SAVE_COLOR(6)
    SAVE_COLOR(7)

#undef SAVE_COLOR
}
#endif // ENABLE_TONEMAPX_AVX_INTRINSICS

X86_64_V3 void tonemap_frame_420p10_2_420p_avx(uint8_t *dsty, uint8_t *dstu, uint8_t *dstv,
                                               const uint16_t *srcy, const uint16_t *srcu, const uint16_t *srcv,
                                               const int *dstlinesize, const int *srclinesize,
                                               int dstdepth, int srcdepth,
                                               int width, int height,
                                               const struct TonemapIntParams *params)
{
#ifdef ENABLE_TONEMAPX_AVX_INTRINSICS
    uint8_t *rdsty = dsty;
    uint8_t *rdstu = dstu;
    uint8_t *rdstv = dstv;
    const uint16_t *rsrcy = srcy;
    const uint16_t *rsrcu = srcu;
    const uint16_t *rsrcv = srcv;
    int rheight = height;
    // not zero when not divisible by 16
    // intentionally leave last pixel emtpy when input is odd
    int remainw = width & 14;

    const int in_depth = srcdepth;
    const int in_uv_offset = 128 << (in_depth - 8);
    const int in_sh = in_depth - 1;
    const int in_rnd = 1 << (in_sh - 1);

    const int out_depth = dstdepth;
    const int out_uv_offset = 128 << (out_depth - 8);
    const int out_sh = 29 - out_depth;
    const int out_rnd = 1 << (out_sh - 1);

    int cy  = (*params->yuv2rgb_coeffs)[0][0][0];
    int crv = (*params->yuv2rgb_coeffs)[0][2][0];
    int cgu = (*params->yuv2rgb_coeffs)[1][1][0];
    int cgv = (*params->yuv2rgb_coeffs)[1][2][0];
    int cbu = (*params->yuv2rgb_coeffs)[2][1][0];

    int cry   = (*params->rgb2yuv_coeffs)[0][0][0];
    int cgy   = (*params->rgb2yuv_coeffs)[0][1][0];
    int cby   = (*params->rgb2yuv_coeffs)[0][2][0];
    int cru   = (*params->rgb2yuv_coeffs)[1][0][0];
    int ocgu  = (*params->rgb2yuv_coeffs)[1][1][0];
    int cburv = (*params->rgb2yuv_coeffs)[1][2][0];
    int ocgv  = (*params->rgb2yuv_coeffs)[2][1][0];
    int cbv   = (*params->rgb2yuv_coeffs)[2][2][0];

    int16_t r[16], g[16], b[16];
    int16_t r1[16], g1[16], b1[16];
    __m256i in_yuv_offx8 = _mm256_set1_epi32(params->in_yuv_off);
    __m256i in_uv_offx8 = _mm256_set1_epi32(in_uv_offset);
    __m256i cyx8 = _mm256_set1_epi32(cy);
    __m256i rndx8 = _mm256_set1_epi32(in_rnd);

    __m256i ux8, vx8;
    __m256i y0x16, y1x16;
    __m256i y0x8a, y0x8b, y1x8a, y1x8b, ux8a, ux8b, vx8a, vx8b;
    __m256i r0x8a, g0x8a, b0x8a, r0x8b, g0x8b, b0x8b;
    __m256i r1x8a, g1x8a, b1x8a, r1x8b, g1x8b, b1x8b;

    __m256i r0ox16, g0ox16, b0ox16;
    __m256i y0ox16;
    __m256i roax8, robx8, goax8, gobx8, boax8, bobx8;
    __m256i yoax8, yobx8;

    __m256i r1ox16, g1ox16, b1ox16;
    __m256i y1ox16;
    __m256i r1oax8, r1obx8, g1oax8, g1obx8, b1oax8, b1obx8;
    __m256i y1oax8, y1obx8;
    __m256i uox8, vox8, ravgx8, gavgx8, bavgx8;
    for (; height > 1; height -= 2,
                       dsty += dstlinesize[0] * 2, dstu += dstlinesize[1], dstv += dstlinesize[2],
                       srcy += srclinesize[0], srcu += srclinesize[1] / 2, srcv += srclinesize[2] / 2) {
        for (int xx = 0; xx < width >> 4; xx++) {
            int x = xx << 4;

            y0x16 = _mm256_lddqu_si256((__m256i*)(srcy + x));
            y1x16 = _mm256_lddqu_si256((__m256i*)(srcy + (srclinesize[0] / 2 + x)));
            ux8 = _mm256_cvtepi16_epi32(_mm_lddqu_si128((__m128i_u *)(srcu + (x >> 1))));
            vx8 = _mm256_cvtepi16_epi32(_mm_lddqu_si128((__m128i_u *)(srcv + (x >> 1))));

            y0x8a = _mm256_cvtepi16_epi32(_mm256_extracti128_si256(y0x16, 0));
            y0x8b = _mm256_cvtepi16_epi32(_mm256_extracti128_si256(y0x16, 1));
            y1x8a = _mm256_cvtepi16_epi32(_mm256_extracti128_si256(y1x16, 0));
            y1x8b = _mm256_cvtepi16_epi32(_mm256_extracti128_si256(y1x16, 1));

            y0x8a = _mm256_sub_epi32(y0x8a, in_yuv_offx8);
            y1x8a = _mm256_sub_epi32(y1x8a, in_yuv_offx8);
            y0x8b = _mm256_sub_epi32(y0x8b, in_yuv_offx8);
            y1x8b = _mm256_sub_epi32(y1x8b, in_yuv_offx8);
            ux8 = _mm256_sub_epi32(ux8, in_uv_offx8);
            vx8 = _mm256_sub_epi32(vx8, in_uv_offx8);

            ux8a = _mm256_permutevar8x32_epi32(ux8, _mm256_set_epi32(3, 3, 2, 2, 1, 1, 0, 0));
            ux8b = _mm256_permutevar8x32_epi32(ux8, _mm256_set_epi32(7, 7, 6, 6, 5, 5, 4, 4));
            vx8a = _mm256_permutevar8x32_epi32(vx8, _mm256_set_epi32(3, 3, 2, 2, 1, 1, 0, 0));
            vx8b = _mm256_permutevar8x32_epi32(vx8, _mm256_set_epi32(7, 7, 6, 6, 5, 5, 4, 4));

            // r = av_clip_int16((y * cy + crv * v + in_rnd) >> in_sh);
            r0x8a = g0x8a = b0x8a = _mm256_mullo_epi32(y0x8a, cyx8);
            r0x8a = _mm256_add_epi32(r0x8a, _mm256_mullo_epi32(vx8a, _mm256_set1_epi32(crv)));
            r0x8a = _mm256_add_epi32(r0x8a, rndx8);
            r0x8a = _mm256_srai_epi32(r0x8a, in_sh);
            r0x8a = av_clip_int16_avx(r0x8a);

            r1x8a = g1x8a = b1x8a = _mm256_mullo_epi32(y1x8a, cyx8);
            r1x8a = _mm256_add_epi32(r1x8a, _mm256_mullo_epi32(vx8a, _mm256_set1_epi32(crv)));
            r1x8a = _mm256_add_epi32(r1x8a, rndx8);
            r1x8a = _mm256_srai_epi32(r1x8a, in_sh);
            r1x8a = av_clip_int16_avx(r1x8a);

            // g = av_clip_int16((y * cy + cgu * u + cgv * v + in_rnd) >> in_sh);
            g0x8a = _mm256_add_epi32(g0x8a, _mm256_mullo_epi32(ux8a, _mm256_set1_epi32(cgu)));
            g0x8a = _mm256_add_epi32(g0x8a, _mm256_mullo_epi32(vx8a, _mm256_set1_epi32(cgv)));
            g0x8a = _mm256_add_epi32(g0x8a, rndx8);
            g0x8a = _mm256_srai_epi32(g0x8a, in_sh);
            g0x8a = av_clip_int16_avx(g0x8a);

            g1x8a = _mm256_add_epi32(g1x8a, _mm256_mullo_epi32(ux8a, _mm256_set1_epi32(cgu)));
            g1x8a = _mm256_add_epi32(g1x8a, _mm256_mullo_epi32(vx8a, _mm256_set1_epi32(cgv)));
            g1x8a = _mm256_add_epi32(g1x8a, rndx8);
            g1x8a = _mm256_srai_epi32(g1x8a, in_sh);
            g1x8a = av_clip_int16_avx(g1x8a);

            // b = av_clip_int16((y * cy + cbu * u + in_rnd) >> in_sh);
            b0x8a = _mm256_add_epi32(b0x8a, _mm256_mullo_epi32(ux8a, _mm256_set1_epi32(cbu)));
            b0x8a = _mm256_add_epi32(b0x8a, rndx8);
            b0x8a = _mm256_srai_epi32(b0x8a, in_sh);
            b0x8a = av_clip_int16_avx(b0x8a);

            b1x8a = _mm256_add_epi32(b1x8a, _mm256_mullo_epi32(ux8a, _mm256_set1_epi32(cbu)));
            b1x8a = _mm256_add_epi32(b1x8a, rndx8);
            b1x8a = _mm256_srai_epi32(b1x8a, in_sh);
            b1x8a = av_clip_int16_avx(b1x8a);

            r0x8b = g0x8b = b0x8b = _mm256_mullo_epi32(y0x8b, cyx8);
            r0x8b = _mm256_add_epi32(r0x8b, _mm256_mullo_epi32(vx8b, _mm256_set1_epi32(crv)));
            r0x8b = _mm256_add_epi32(r0x8b, rndx8);
            r0x8b = _mm256_srai_epi32(r0x8b, in_sh);
            r0x8b = av_clip_int16_avx(r0x8b);

            r1x8b = g1x8b = b1x8b = _mm256_mullo_epi32(y1x8b, cyx8);
            r1x8b = _mm256_add_epi32(r1x8b, _mm256_mullo_epi32(vx8b, _mm256_set1_epi32(crv)));
            r1x8b = _mm256_add_epi32(r1x8b, rndx8);
            r1x8b = _mm256_srai_epi32(r1x8b, in_sh);
            r1x8b = av_clip_int16_avx(r1x8b);

            g0x8b = _mm256_add_epi32(g0x8b, _mm256_mullo_epi32(ux8b, _mm256_set1_epi32(cgu)));
            g0x8b = _mm256_add_epi32(g0x8b, _mm256_mullo_epi32(vx8b, _mm256_set1_epi32(cgv)));
            g0x8b = _mm256_add_epi32(g0x8b, rndx8);
            g0x8b = _mm256_srai_epi32(g0x8b, in_sh);
            g0x8b = av_clip_int16_avx(g0x8b);

            g1x8b = _mm256_add_epi32(g1x8b, _mm256_mullo_epi32(ux8b, _mm256_set1_epi32(cgu)));
            g1x8b = _mm256_add_epi32(g1x8b, _mm256_mullo_epi32(vx8b, _mm256_set1_epi32(cgv)));
            g1x8b = _mm256_add_epi32(g1x8b, rndx8);
            g1x8b = _mm256_srai_epi32(g1x8b, in_sh);
            g1x8b = av_clip_int16_avx(g1x8b);

            b0x8b = _mm256_add_epi32(b0x8b, _mm256_mullo_epi32(ux8b, _mm256_set1_epi32(cbu)));
            b0x8b = _mm256_add_epi32(b0x8b, rndx8);
            b0x8b = _mm256_srai_epi32(b0x8b, in_sh);
            b0x8b = av_clip_int16_avx(b0x8b);

            b1x8b = _mm256_add_epi32(b1x8b, _mm256_mullo_epi32(ux8b, _mm256_set1_epi32(cbu)));
            b1x8b = _mm256_add_epi32(b1x8b, rndx8);
            b1x8b = _mm256_srai_epi32(b1x8b, in_sh);
            b1x8b = av_clip_int16_avx(b1x8b);

            tonemap_int32x8_avx(r0x8a, g0x8a, b0x8a, r, g, b,
                                params->lin_lut, params->tonemap_lut, params->delin_lut,
                                params->coeffs, params->ocoeffs, params->desat, params->rgb2rgb_coeffs,
                                params->rgb2rgb_passthrough);
            tonemap_int32x8_avx(r1x8a, g1x8a, b1x8a, r1, g1, b1,
                                params->lin_lut, params->tonemap_lut, params->delin_lut,
                                params->coeffs, params->ocoeffs, params->desat, params->rgb2rgb_coeffs,
                                params->rgb2rgb_passthrough);
            tonemap_int32x8_avx(r0x8b, g0x8b, b0x8b, &r[8], &g[8], &b[8],
                                params->lin_lut, params->tonemap_lut, params->delin_lut,
                                params->coeffs, params->ocoeffs, params->desat, params->rgb2rgb_coeffs,
                                params->rgb2rgb_passthrough);
            tonemap_int32x8_avx(r1x8b, g1x8b, b1x8b, &r1[8], &g1[8], &b1[8],
                                params->lin_lut, params->tonemap_lut, params->delin_lut,
                                params->coeffs, params->ocoeffs, params->desat, params->rgb2rgb_coeffs,
                                params->rgb2rgb_passthrough);

            r0ox16 = _mm256_lddqu_si256((const __m256i_u *)r);
            g0ox16 = _mm256_lddqu_si256((const __m256i_u *)g);
            b0ox16 = _mm256_lddqu_si256((const __m256i_u *)b);

            roax8 = _mm256_cvtepi16_epi32(_mm256_extracti128_si256(r0ox16, 0));
            goax8 = _mm256_cvtepi16_epi32(_mm256_extracti128_si256(g0ox16, 0));
            boax8 = _mm256_cvtepi16_epi32(_mm256_extracti128_si256(b0ox16, 0));

            robx8 = _mm256_cvtepi16_epi32(_mm256_extracti128_si256(r0ox16, 1));
            gobx8 = _mm256_cvtepi16_epi32(_mm256_extracti128_si256(g0ox16, 1));
            bobx8 = _mm256_cvtepi16_epi32(_mm256_extracti128_si256(b0ox16, 1));

            yoax8 = _mm256_mullo_epi32(roax8, _mm256_set1_epi32(cry));
            yoax8 = _mm256_add_epi32(yoax8, _mm256_mullo_epi32(goax8, _mm256_set1_epi32(cgy)));
            yoax8 = _mm256_add_epi32(yoax8, _mm256_mullo_epi32(boax8, _mm256_set1_epi32(cby)));
            yoax8 = _mm256_add_epi32(yoax8, _mm256_set1_epi32(out_rnd));
            yoax8 = _mm256_srai_epi32(yoax8, out_sh);
            yoax8 = _mm256_add_epi32(yoax8, _mm256_set1_epi32(params->out_yuv_off));

            yobx8 = _mm256_mullo_epi32(robx8, _mm256_set1_epi32(cry));
            yobx8 = _mm256_add_epi32(yobx8, _mm256_mullo_epi32(gobx8, _mm256_set1_epi32(cgy)));
            yobx8 = _mm256_add_epi32(yobx8, _mm256_mullo_epi32(bobx8, _mm256_set1_epi32(cby)));
            yobx8 = _mm256_add_epi32(yobx8, _mm256_set1_epi32(out_rnd));
            yobx8 = _mm256_srai_epi32(yobx8, out_sh);
            yobx8 = _mm256_add_epi32(yobx8, _mm256_set1_epi32(params->out_yuv_off));

            y0ox16 = _mm256_packs_epi32(yoax8, yobx8);
            y0ox16 = _mm256_permute4x64_epi64(y0ox16, _MM_SHUFFLE(3, 1, 2, 0));
            _mm_storeu_si128((__m128i_u *) &dsty[x], _mm256_castsi256_si128(_mm256_permute4x64_epi64(_mm256_packus_epi16(y0ox16, _mm256_setzero_si256()), _MM_SHUFFLE(3, 1, 2, 0))));

            r1ox16 = _mm256_lddqu_si256((const __m256i_u *)r1);
            g1ox16 = _mm256_lddqu_si256((const __m256i_u *)g1);
            b1ox16 = _mm256_lddqu_si256((const __m256i_u *)b1);

            r1oax8 = _mm256_cvtepi16_epi32(_mm256_extracti128_si256(r1ox16, 0));
            g1oax8 = _mm256_cvtepi16_epi32(_mm256_extracti128_si256(g1ox16, 0));
            b1oax8 = _mm256_cvtepi16_epi32(_mm256_extracti128_si256(b1ox16, 0));

            r1obx8 = _mm256_cvtepi16_epi32(_mm256_extracti128_si256(r1ox16, 1));
            g1obx8 = _mm256_cvtepi16_epi32(_mm256_extracti128_si256(g1ox16, 1));
            b1obx8 = _mm256_cvtepi16_epi32(_mm256_extracti128_si256(b1ox16, 1));

            y1oax8 = _mm256_mullo_epi32(r1oax8, _mm256_set1_epi32(cry));
            y1oax8 = _mm256_add_epi32(y1oax8, _mm256_mullo_epi32(g1oax8, _mm256_set1_epi32(cgy)));
            y1oax8 = _mm256_add_epi32(y1oax8, _mm256_mullo_epi32(b1oax8, _mm256_set1_epi32(cby)));
            y1oax8 = _mm256_add_epi32(y1oax8, _mm256_set1_epi32(out_rnd));
            y1oax8 = _mm256_srai_epi32(y1oax8, out_sh);
            y1oax8 = _mm256_add_epi32(y1oax8, _mm256_set1_epi32(params->out_yuv_off));

            y1obx8 = _mm256_mullo_epi32(r1obx8, _mm256_set1_epi32(cry));
            y1obx8 = _mm256_add_epi32(y1obx8, _mm256_mullo_epi32(g1obx8, _mm256_set1_epi32(cgy)));
            y1obx8 = _mm256_add_epi32(y1obx8, _mm256_mullo_epi32(b1obx8, _mm256_set1_epi32(cby)));
            y1obx8 = _mm256_add_epi32(y1obx8, _mm256_set1_epi32(out_rnd));
            y1obx8 = _mm256_srai_epi32(y1obx8, out_sh);
            y1obx8 = _mm256_add_epi32(y1obx8, _mm256_set1_epi32(params->out_yuv_off));

            y1ox16 = _mm256_packs_epi32(y1oax8, y1obx8);
            y1ox16 = _mm256_permute4x64_epi64(y1ox16, _MM_SHUFFLE(3, 1, 2, 0));
            _mm_storeu_si128((__m128i_u *) &dsty[x + dstlinesize[0]], _mm256_castsi256_si128(_mm256_permute4x64_epi64(_mm256_packus_epi16(y1ox16, _mm256_setzero_si256()), _MM_SHUFFLE(3, 1, 2, 0))));

            ravgx8 = _mm256_hadd_epi32(roax8, robx8);
            ravgx8 = _mm256_add_epi32(ravgx8, _mm256_hadd_epi32(r1oax8, r1obx8));
            ravgx8 = _mm256_permute4x64_epi64(ravgx8, _MM_SHUFFLE(3, 1, 2, 0));
            ravgx8 = _mm256_add_epi32(ravgx8, _mm256_set1_epi32(2));
            ravgx8 = _mm256_srai_epi32(ravgx8, 2);

            gavgx8 = _mm256_hadd_epi32(goax8, gobx8);
            gavgx8 = _mm256_add_epi32(gavgx8, _mm256_hadd_epi32(g1oax8, g1obx8));
            gavgx8 = _mm256_permute4x64_epi64(gavgx8, _MM_SHUFFLE(3, 1, 2, 0));
            gavgx8 = _mm256_add_epi32(gavgx8, _mm256_set1_epi32(2));
            gavgx8 = _mm256_srai_epi32(gavgx8, 2);

            bavgx8 = _mm256_hadd_epi32(boax8, bobx8);
            bavgx8 = _mm256_add_epi32(bavgx8, _mm256_hadd_epi32(b1oax8, b1obx8));
            bavgx8 = _mm256_permute4x64_epi64(bavgx8, _MM_SHUFFLE(3, 1, 2, 0));
            bavgx8 = _mm256_add_epi32(bavgx8, _mm256_set1_epi32(2));
            bavgx8 = _mm256_srai_epi32(bavgx8, 2);

            uox8 = _mm256_add_epi32(_mm256_set1_epi32(out_rnd), _mm256_mullo_epi32(ravgx8, _mm256_set1_epi32(cru)));
            uox8 = _mm256_add_epi32(uox8, _mm256_mullo_epi32(gavgx8, _mm256_set1_epi32(ocgu)));
            uox8 = _mm256_add_epi32(uox8, _mm256_mullo_epi32(bavgx8, _mm256_set1_epi32(cburv)));
            uox8 = _mm256_srai_epi32(uox8, out_sh);
            uox8 = _mm256_add_epi32(uox8, _mm256_set1_epi32(out_uv_offset));
            uox8 = _mm256_packs_epi32(uox8, _mm256_setzero_si256());
            uox8 = _mm256_permute4x64_epi64(uox8, _MM_SHUFFLE(3, 1, 2, 0));
            uox8 = _mm256_packus_epi16(uox8, _mm256_setzero_si256());
            _mm_storeu_si64(&dstu[x >> 1], _mm256_castsi256_si128(uox8));

            vox8 = _mm256_add_epi32(_mm256_set1_epi32(out_rnd), _mm256_mullo_epi32(ravgx8, _mm256_set1_epi32(cburv)));
            vox8 = _mm256_add_epi32(vox8, _mm256_mullo_epi32(gavgx8, _mm256_set1_epi32(ocgv)));
            vox8 = _mm256_add_epi32(vox8, _mm256_mullo_epi32(bavgx8, _mm256_set1_epi32(cbv)));
            vox8 = _mm256_srai_epi32(vox8, out_sh);
            vox8 = _mm256_add_epi32(vox8, _mm256_set1_epi32(out_uv_offset));
            vox8 = _mm256_packs_epi32(vox8, _mm256_setzero_si256());
            vox8 = _mm256_permute4x64_epi64(vox8, _MM_SHUFFLE(3, 1, 2, 0));
            vox8 = _mm256_packus_epi16(vox8, _mm256_setzero_si256());
            _mm_storeu_si64(&dstv[x >> 1], _mm256_castsi256_si128(vox8));
        }
    }

    // Process remaining pixels cannot fill the full simd register with scalar version
    if (remainw) {
        int offset = width & (int)0xfffffff0;
        rdsty += offset;
        rdstu += offset >> 1;
        rdstv += offset >> 1;
        rsrcy += offset;
        rsrcu += offset >> 1;
        rsrcv += offset >> 1;
        tonemap_frame_420p10_2_420p(rdsty, rdstu, rdstv,
                                    rsrcy, rsrcu, rsrcv,
                                    dstlinesize, srclinesize,
                                    dstdepth, srcdepth,
                                    remainw, rheight, params);
    }
#endif // ENABLE_TONEMAPX_AVX_INTRINSICS
}

X86_64_V3 void tonemap_frame_420p10_2_420p10_avx(uint16_t *dsty, uint16_t *dstu, uint16_t *dstv,
                                                 const uint16_t *srcy, const uint16_t *srcu, const uint16_t *srcv,
                                                 const int *dstlinesize, const int *srclinesize,
                                                 int dstdepth, int srcdepth,
                                                 int width, int height,
                                                 const struct TonemapIntParams *params)
{
#ifdef ENABLE_TONEMAPX_AVX_INTRINSICS
    uint16_t *rdsty = dsty;
    uint16_t *rdstu = dstu;
    uint16_t *rdstv = dstv;
    const uint16_t *rsrcy = srcy;
    const uint16_t *rsrcu = srcu;
    const uint16_t *rsrcv = srcv;
    int rheight = height;
    // not zero when not divisible by 8
    // intentionally leave last pixel emtpy when input is odd
    int remainw = width & 14;

    const int in_depth = srcdepth;
    const int in_uv_offset = 128 << (in_depth - 8);
    const int in_sh = in_depth - 1;
    const int in_rnd = 1 << (in_sh - 1);

    const int out_depth = dstdepth;
    const int out_uv_offset = 128 << (out_depth - 8);
    const int out_sh = 29 - out_depth;
    const int out_rnd = 1 << (out_sh - 1);

    int cy  = (*params->yuv2rgb_coeffs)[0][0][0];
    int crv = (*params->yuv2rgb_coeffs)[0][2][0];
    int cgu = (*params->yuv2rgb_coeffs)[1][1][0];
    int cgv = (*params->yuv2rgb_coeffs)[1][2][0];
    int cbu = (*params->yuv2rgb_coeffs)[2][1][0];

    int cry   = (*params->rgb2yuv_coeffs)[0][0][0];
    int cgy   = (*params->rgb2yuv_coeffs)[0][1][0];
    int cby   = (*params->rgb2yuv_coeffs)[0][2][0];
    int cru   = (*params->rgb2yuv_coeffs)[1][0][0];
    int ocgu  = (*params->rgb2yuv_coeffs)[1][1][0];
    int cburv = (*params->rgb2yuv_coeffs)[1][2][0];
    int ocgv  = (*params->rgb2yuv_coeffs)[2][1][0];
    int cbv   = (*params->rgb2yuv_coeffs)[2][2][0];

    int16_t r[16], g[16], b[16];
    int16_t r1[16], g1[16], b1[16];
    __m256i in_yuv_offx8 = _mm256_set1_epi32(params->in_yuv_off);
    __m256i in_uv_offx8 = _mm256_set1_epi32(in_uv_offset);
    __m256i cyx8 = _mm256_set1_epi32(cy);
    __m256i rndx8 = _mm256_set1_epi32(in_rnd);

    __m256i r0ox16, g0ox16, b0ox16;
    __m256i y0ox16;
    __m256i roax8, robx8, goax8, gobx8, boax8, bobx8;
    __m256i yoax8, yobx8;
    __m256i ux8, vx8;
    __m256i y0x16, y1x16;
    __m256i y0x8a, y0x8b, y1x8a, y1x8b, ux8a, ux8b, vx8a, vx8b;
    __m256i r0x8a, g0x8a, b0x8a, r0x8b, g0x8b, b0x8b;
    __m256i r1x8a, g1x8a, b1x8a, r1x8b, g1x8b, b1x8b;

    __m256i r1ox16, g1ox16, b1ox16;
    __m256i y1ox16;
    __m256i r1oax8, r1obx8, g1oax8, g1obx8, b1oax8, b1obx8;
    __m256i y1oax8, y1obx8;
    __m256i uox8, vox8, ravgx8, gavgx8, bavgx8;
    for (; height > 1; height -= 2,
                       dsty += dstlinesize[0], dstu += dstlinesize[1] / 2, dstv += dstlinesize[1] / 2,
                       srcy += srclinesize[0], srcu += srclinesize[1] / 2, srcv += srclinesize[1] / 2) {
        for (int xx = 0; xx < width >> 4; xx++) {
            int x = xx << 4;

            y0x16 = _mm256_lddqu_si256((__m256i*)(srcy + x));
            y1x16 = _mm256_lddqu_si256((__m256i*)(srcy + (srclinesize[0] / 2 + x)));
            ux8 = _mm256_cvtepi16_epi32(_mm_lddqu_si128((__m128i_u *)(srcu + (x >> 1))));
            vx8 = _mm256_cvtepi16_epi32(_mm_lddqu_si128((__m128i_u *)(srcv + (x >> 1))));

            y0x8a = _mm256_cvtepi16_epi32(_mm256_extracti128_si256(y0x16, 0));
            y0x8b = _mm256_cvtepi16_epi32(_mm256_extracti128_si256(y0x16, 1));
            y1x8a = _mm256_cvtepi16_epi32(_mm256_extracti128_si256(y1x16, 0));
            y1x8b = _mm256_cvtepi16_epi32(_mm256_extracti128_si256(y1x16, 1));

            y0x8a = _mm256_sub_epi32(y0x8a, in_yuv_offx8);
            y1x8a = _mm256_sub_epi32(y1x8a, in_yuv_offx8);
            y0x8b = _mm256_sub_epi32(y0x8b, in_yuv_offx8);
            y1x8b = _mm256_sub_epi32(y1x8b, in_yuv_offx8);
            ux8 = _mm256_sub_epi32(ux8, in_uv_offx8);
            vx8 = _mm256_sub_epi32(vx8, in_uv_offx8);

            ux8a = _mm256_permutevar8x32_epi32(ux8, _mm256_set_epi32(3, 3, 2, 2, 1, 1, 0, 0));
            ux8b = _mm256_permutevar8x32_epi32(ux8, _mm256_set_epi32(7, 7, 6, 6, 5, 5, 4, 4));
            vx8a = _mm256_permutevar8x32_epi32(vx8, _mm256_set_epi32(3, 3, 2, 2, 1, 1, 0, 0));
            vx8b = _mm256_permutevar8x32_epi32(vx8, _mm256_set_epi32(7, 7, 6, 6, 5, 5, 4, 4));

            // r = av_clip_int16((y * cy + crv * v + in_rnd) >> in_sh);
            r0x8a = g0x8a = b0x8a = _mm256_mullo_epi32(y0x8a, cyx8);
            r0x8a = _mm256_add_epi32(r0x8a, _mm256_mullo_epi32(vx8a, _mm256_set1_epi32(crv)));
            r0x8a = _mm256_add_epi32(r0x8a, rndx8);
            r0x8a = _mm256_srai_epi32(r0x8a, in_sh);
            r0x8a = av_clip_int16_avx(r0x8a);

            r1x8a = g1x8a = b1x8a = _mm256_mullo_epi32(y1x8a, cyx8);
            r1x8a = _mm256_add_epi32(r1x8a, _mm256_mullo_epi32(vx8a, _mm256_set1_epi32(crv)));
            r1x8a = _mm256_add_epi32(r1x8a, rndx8);
            r1x8a = _mm256_srai_epi32(r1x8a, in_sh);
            r1x8a = av_clip_int16_avx(r1x8a);

            // g = av_clip_int16((y * cy + cgu * u + cgv * v + in_rnd) >> in_sh);
            g0x8a = _mm256_add_epi32(g0x8a, _mm256_mullo_epi32(ux8a, _mm256_set1_epi32(cgu)));
            g0x8a = _mm256_add_epi32(g0x8a, _mm256_mullo_epi32(vx8a, _mm256_set1_epi32(cgv)));
            g0x8a = _mm256_add_epi32(g0x8a, rndx8);
            g0x8a = _mm256_srai_epi32(g0x8a, in_sh);
            g0x8a = av_clip_int16_avx(g0x8a);

            g1x8a = _mm256_add_epi32(g1x8a, _mm256_mullo_epi32(ux8a, _mm256_set1_epi32(cgu)));
            g1x8a = _mm256_add_epi32(g1x8a, _mm256_mullo_epi32(vx8a, _mm256_set1_epi32(cgv)));
            g1x8a = _mm256_add_epi32(g1x8a, rndx8);
            g1x8a = _mm256_srai_epi32(g1x8a, in_sh);
            g1x8a = av_clip_int16_avx(g1x8a);

            // b = av_clip_int16((y * cy + cbu * u + in_rnd) >> in_sh);
            b0x8a = _mm256_add_epi32(b0x8a, _mm256_mullo_epi32(ux8a, _mm256_set1_epi32(cbu)));
            b0x8a = _mm256_add_epi32(b0x8a, rndx8);
            b0x8a = _mm256_srai_epi32(b0x8a, in_sh);
            b0x8a = av_clip_int16_avx(b0x8a);

            b1x8a = _mm256_add_epi32(b1x8a, _mm256_mullo_epi32(ux8a, _mm256_set1_epi32(cbu)));
            b1x8a = _mm256_add_epi32(b1x8a, rndx8);
            b1x8a = _mm256_srai_epi32(b1x8a, in_sh);
            b1x8a = av_clip_int16_avx(b1x8a);

            r0x8b = g0x8b = b0x8b = _mm256_mullo_epi32(y0x8b, cyx8);
            r0x8b = _mm256_add_epi32(r0x8b, _mm256_mullo_epi32(vx8b, _mm256_set1_epi32(crv)));
            r0x8b = _mm256_add_epi32(r0x8b, rndx8);
            r0x8b = _mm256_srai_epi32(r0x8b, in_sh);
            r0x8b = av_clip_int16_avx(r0x8b);

            r1x8b = g1x8b = b1x8b = _mm256_mullo_epi32(y1x8b, cyx8);
            r1x8b = _mm256_add_epi32(r1x8b, _mm256_mullo_epi32(vx8b, _mm256_set1_epi32(crv)));
            r1x8b = _mm256_add_epi32(r1x8b, rndx8);
            r1x8b = _mm256_srai_epi32(r1x8b, in_sh);
            r1x8b = av_clip_int16_avx(r1x8b);

            g0x8b = _mm256_add_epi32(g0x8b, _mm256_mullo_epi32(ux8b, _mm256_set1_epi32(cgu)));
            g0x8b = _mm256_add_epi32(g0x8b, _mm256_mullo_epi32(vx8b, _mm256_set1_epi32(cgv)));
            g0x8b = _mm256_add_epi32(g0x8b, rndx8);
            g0x8b = _mm256_srai_epi32(g0x8b, in_sh);
            g0x8b = av_clip_int16_avx(g0x8b);

            g1x8b = _mm256_add_epi32(g1x8b, _mm256_mullo_epi32(ux8b, _mm256_set1_epi32(cgu)));
            g1x8b = _mm256_add_epi32(g1x8b, _mm256_mullo_epi32(vx8b, _mm256_set1_epi32(cgv)));
            g1x8b = _mm256_add_epi32(g1x8b, rndx8);
            g1x8b = _mm256_srai_epi32(g1x8b, in_sh);
            g1x8b = av_clip_int16_avx(g1x8b);

            b0x8b = _mm256_add_epi32(b0x8b, _mm256_mullo_epi32(ux8b, _mm256_set1_epi32(cbu)));
            b0x8b = _mm256_add_epi32(b0x8b, rndx8);
            b0x8b = _mm256_srai_epi32(b0x8b, in_sh);
            b0x8b = av_clip_int16_avx(b0x8b);

            b1x8b = _mm256_add_epi32(b1x8b, _mm256_mullo_epi32(ux8b, _mm256_set1_epi32(cbu)));
            b1x8b = _mm256_add_epi32(b1x8b, rndx8);
            b1x8b = _mm256_srai_epi32(b1x8b, in_sh);
            b1x8b = av_clip_int16_avx(b1x8b);

            tonemap_int32x8_avx(r0x8a, g0x8a, b0x8a, r, g, b,
                                params->lin_lut, params->tonemap_lut, params->delin_lut,
                                params->coeffs, params->ocoeffs, params->desat, params->rgb2rgb_coeffs,
                                params->rgb2rgb_passthrough);
            tonemap_int32x8_avx(r1x8a, g1x8a, b1x8a, r1, g1, b1,
                                params->lin_lut, params->tonemap_lut, params->delin_lut,
                                params->coeffs, params->ocoeffs, params->desat, params->rgb2rgb_coeffs,
                                params->rgb2rgb_passthrough);
            tonemap_int32x8_avx(r0x8b, g0x8b, b0x8b, &r[8], &g[8], &b[8],
                                params->lin_lut, params->tonemap_lut, params->delin_lut,
                                params->coeffs, params->ocoeffs, params->desat, params->rgb2rgb_coeffs,
                                params->rgb2rgb_passthrough);
            tonemap_int32x8_avx(r1x8b, g1x8b, b1x8b, &r1[8], &g1[8], &b1[8],
                                params->lin_lut, params->tonemap_lut, params->delin_lut,
                                params->coeffs, params->ocoeffs, params->desat, params->rgb2rgb_coeffs,
                                params->rgb2rgb_passthrough);

            r0ox16 = _mm256_lddqu_si256((const __m256i_u *)r);
            g0ox16 = _mm256_lddqu_si256((const __m256i_u *)g);
            b0ox16 = _mm256_lddqu_si256((const __m256i_u *)b);

            roax8 = _mm256_cvtepi16_epi32(_mm256_extracti128_si256(r0ox16, 0));
            goax8 = _mm256_cvtepi16_epi32(_mm256_extracti128_si256(g0ox16, 0));
            boax8 = _mm256_cvtepi16_epi32(_mm256_extracti128_si256(b0ox16, 0));

            robx8 = _mm256_cvtepi16_epi32(_mm256_extracti128_si256(r0ox16, 1));
            gobx8 = _mm256_cvtepi16_epi32(_mm256_extracti128_si256(g0ox16, 1));
            bobx8 = _mm256_cvtepi16_epi32(_mm256_extracti128_si256(b0ox16, 1));

            yoax8 = _mm256_mullo_epi32(roax8, _mm256_set1_epi32(cry));
            yoax8 = _mm256_add_epi32(yoax8, _mm256_mullo_epi32(goax8, _mm256_set1_epi32(cgy)));
            yoax8 = _mm256_add_epi32(yoax8, _mm256_mullo_epi32(boax8, _mm256_set1_epi32(cby)));
            yoax8 = _mm256_add_epi32(yoax8, _mm256_set1_epi32(out_rnd));
            yoax8 = _mm256_srai_epi32(yoax8, out_sh);
            yoax8 = _mm256_add_epi32(yoax8, _mm256_set1_epi32(params->out_yuv_off));

            yobx8 = _mm256_mullo_epi32(robx8, _mm256_set1_epi32(cry));
            yobx8 = _mm256_add_epi32(yobx8, _mm256_mullo_epi32(gobx8, _mm256_set1_epi32(cgy)));
            yobx8 = _mm256_add_epi32(yobx8, _mm256_mullo_epi32(bobx8, _mm256_set1_epi32(cby)));
            yobx8 = _mm256_add_epi32(yobx8, _mm256_set1_epi32(out_rnd));
            yobx8 = _mm256_srai_epi32(yobx8, out_sh);
            yobx8 = _mm256_add_epi32(yobx8, _mm256_set1_epi32(params->out_yuv_off));

            y0ox16 = _mm256_packus_epi32(yoax8, yobx8);
            y0ox16 = _mm256_permute4x64_epi64(y0ox16, _MM_SHUFFLE(3, 1, 2, 0));
            _mm256_storeu_si256((__m256i_u *) &dsty[x], y0ox16);

            r1ox16 = _mm256_lddqu_si256((const __m256i_u *)r1);
            g1ox16 = _mm256_lddqu_si256((const __m256i_u *)g1);
            b1ox16 = _mm256_lddqu_si256((const __m256i_u *)b1);

            r1oax8 = _mm256_cvtepi16_epi32(_mm256_extracti128_si256(r1ox16, 0));
            g1oax8 = _mm256_cvtepi16_epi32(_mm256_extracti128_si256(g1ox16, 0));
            b1oax8 = _mm256_cvtepi16_epi32(_mm256_extracti128_si256(b1ox16, 0));

            r1obx8 = _mm256_cvtepi16_epi32(_mm256_extracti128_si256(r1ox16, 1));
            g1obx8 = _mm256_cvtepi16_epi32(_mm256_extracti128_si256(g1ox16, 1));
            b1obx8 = _mm256_cvtepi16_epi32(_mm256_extracti128_si256(b1ox16, 1));

            y1oax8 = _mm256_mullo_epi32(r1oax8, _mm256_set1_epi32(cry));
            y1oax8 = _mm256_add_epi32(y1oax8, _mm256_mullo_epi32(g1oax8, _mm256_set1_epi32(cgy)));
            y1oax8 = _mm256_add_epi32(y1oax8, _mm256_mullo_epi32(b1oax8, _mm256_set1_epi32(cby)));
            y1oax8 = _mm256_add_epi32(y1oax8, _mm256_set1_epi32(out_rnd));
            y1oax8 = _mm256_srai_epi32(y1oax8, out_sh);
            y1oax8 = _mm256_add_epi32(y1oax8, _mm256_set1_epi32(params->out_yuv_off));

            y1obx8 = _mm256_mullo_epi32(r1obx8, _mm256_set1_epi32(cry));
            y1obx8 = _mm256_add_epi32(y1obx8, _mm256_mullo_epi32(g1obx8, _mm256_set1_epi32(cgy)));
            y1obx8 = _mm256_add_epi32(y1obx8, _mm256_mullo_epi32(b1obx8, _mm256_set1_epi32(cby)));
            y1obx8 = _mm256_add_epi32(y1obx8, _mm256_set1_epi32(out_rnd));
            y1obx8 = _mm256_srai_epi32(y1obx8, out_sh);
            y1obx8 = _mm256_add_epi32(y1obx8, _mm256_set1_epi32(params->out_yuv_off));

            y1ox16 = _mm256_packus_epi32(y1oax8, y1obx8);
            y1ox16 = _mm256_permute4x64_epi64(y1ox16, _MM_SHUFFLE(3, 1, 2, 0));
            _mm256_storeu_si256((__m256i_u *) &dsty[x + dstlinesize[0] / 2], y1ox16);

            ravgx8 = _mm256_hadd_epi32(roax8, robx8);
            ravgx8 = _mm256_add_epi32(ravgx8, _mm256_hadd_epi32(r1oax8, r1obx8));
            ravgx8 = _mm256_permute4x64_epi64(ravgx8, _MM_SHUFFLE(3, 1, 2, 0));
            ravgx8 = _mm256_add_epi32(ravgx8, _mm256_set1_epi32(2));
            ravgx8 = _mm256_srai_epi32(ravgx8, 2);

            gavgx8 = _mm256_hadd_epi32(goax8, gobx8);
            gavgx8 = _mm256_add_epi32(gavgx8, _mm256_hadd_epi32(g1oax8, g1obx8));
            gavgx8 = _mm256_permute4x64_epi64(gavgx8, _MM_SHUFFLE(3, 1, 2, 0));
            gavgx8 = _mm256_add_epi32(gavgx8, _mm256_set1_epi32(2));
            gavgx8 = _mm256_srai_epi32(gavgx8, 2);

            bavgx8 = _mm256_hadd_epi32(boax8, bobx8);
            bavgx8 = _mm256_add_epi32(bavgx8, _mm256_hadd_epi32(b1oax8, b1obx8));
            bavgx8 = _mm256_permute4x64_epi64(bavgx8, _MM_SHUFFLE(3, 1, 2, 0));
            bavgx8 = _mm256_add_epi32(bavgx8, _mm256_set1_epi32(2));
            bavgx8 = _mm256_srai_epi32(bavgx8, 2);

            uox8 = _mm256_add_epi32(_mm256_set1_epi32(out_rnd), _mm256_mullo_epi32(ravgx8, _mm256_set1_epi32(cru)));
            uox8 = _mm256_add_epi32(uox8, _mm256_mullo_epi32(gavgx8, _mm256_set1_epi32(ocgu)));
            uox8 = _mm256_add_epi32(uox8, _mm256_mullo_epi32(bavgx8, _mm256_set1_epi32(cburv)));
            uox8 = _mm256_srai_epi32(uox8, out_sh);
            uox8 = _mm256_add_epi32(uox8, _mm256_set1_epi32(out_uv_offset));
            uox8 = _mm256_packus_epi32(uox8, _mm256_setzero_si256());
            uox8 = _mm256_permute4x64_epi64(uox8, _MM_SHUFFLE(3, 1, 2, 0));
            _mm_storeu_si128((__m128i_u *) &dstu[x >> 1], _mm256_castsi256_si128(uox8));

            vox8 = _mm256_add_epi32(_mm256_set1_epi32(out_rnd), _mm256_mullo_epi32(ravgx8, _mm256_set1_epi32(cburv)));
            vox8 = _mm256_add_epi32(vox8, _mm256_mullo_epi32(gavgx8, _mm256_set1_epi32(ocgv)));
            vox8 = _mm256_add_epi32(vox8, _mm256_mullo_epi32(bavgx8, _mm256_set1_epi32(cbv)));
            vox8 = _mm256_srai_epi32(vox8, out_sh);
            vox8 = _mm256_add_epi32(vox8, _mm256_set1_epi32(out_uv_offset));
            vox8 = _mm256_packus_epi32(vox8, _mm256_setzero_si256());
            vox8 = _mm256_permute4x64_epi64(vox8, _MM_SHUFFLE(3, 1, 2, 0));
            _mm_storeu_si128((__m128i_u *) &dstv[x >> 1], _mm256_castsi256_si128(vox8));
        }
    }

    // Process remaining pixels cannot fill the full simd register with scalar version
    if (remainw) {
        int offset = width & (int)0xfffffff0;
        rdsty += offset;
        rdstu += offset >> 1;
        rdstv += offset >> 1;
        rsrcy += offset;
        rsrcu += offset >> 1;
        rsrcv += offset >> 1;
        tonemap_frame_420p10_2_420p10(rdsty, rdstu, rdstv,
                                      rsrcy, rsrcu, rsrcv,
                                      dstlinesize, srclinesize,
                                      dstdepth, srcdepth,
                                      remainw, rheight, params);
    }
#endif // ENABLE_TONEMAPX_AVX_INTRINSICS
}

X86_64_V3 void tonemap_frame_p016_p010_2_nv12_avx(uint8_t *dsty, uint8_t *dstuv,
                                                  const uint16_t *srcy, const uint16_t *srcuv,
                                                  const int *dstlinesize, const int *srclinesize,
                                                  int dstdepth, int srcdepth,
                                                  int width, int height,
                                                  const struct TonemapIntParams *params)
{
#ifdef ENABLE_TONEMAPX_AVX_INTRINSICS
    uint8_t *rdsty = dsty;
    uint8_t *rdstuv = dstuv;
    const uint16_t *rsrcy = srcy;
    const uint16_t *rsrcuv = srcuv;
    int rheight = height;
    // not zero when not divisible by 16
    // intentionally leave last pixel emtpy when input is odd
    int remainw = width & 14;

    const int in_depth = srcdepth;
    const int in_uv_offset = 128 << (in_depth - 8);
    const int in_sh = in_depth - 1;
    const int in_rnd = 1 << (in_sh - 1);

    const int out_depth = dstdepth;
    const int out_uv_offset = 128 << (out_depth - 8);
    const int out_sh = 29 - out_depth;
    const int out_rnd = 1 << (out_sh - 1);

    int cy  = (*params->yuv2rgb_coeffs)[0][0][0];
    int crv = (*params->yuv2rgb_coeffs)[0][2][0];
    int cgu = (*params->yuv2rgb_coeffs)[1][1][0];
    int cgv = (*params->yuv2rgb_coeffs)[1][2][0];
    int cbu = (*params->yuv2rgb_coeffs)[2][1][0];

    int cry   = (*params->rgb2yuv_coeffs)[0][0][0];
    int cgy   = (*params->rgb2yuv_coeffs)[0][1][0];
    int cby   = (*params->rgb2yuv_coeffs)[0][2][0];
    int cru   = (*params->rgb2yuv_coeffs)[1][0][0];
    int ocgu  = (*params->rgb2yuv_coeffs)[1][1][0];
    int cburv = (*params->rgb2yuv_coeffs)[1][2][0];
    int ocgv  = (*params->rgb2yuv_coeffs)[2][1][0];
    int cbv   = (*params->rgb2yuv_coeffs)[2][2][0];

    int16_t r[16], g[16], b[16];
    int16_t r1[16], g1[16], b1[16];
    __m256i in_yuv_offx8 = _mm256_set1_epi32(params->in_yuv_off);
    __m256i in_uv_offx8 = _mm256_set1_epi32(in_uv_offset);
    __m256i cyx8 = _mm256_set1_epi32(cy);
    __m256i rndx8 = _mm256_set1_epi32(in_rnd);

    __m256i uvx16, uvx8a, uvx8b;
    __m256i y0x16, y1x16;
    __m256i y0x8a, y0x8b, y1x8a, y1x8b, ux8a, ux8b, vx8a, vx8b;
    __m256i r0x8a, g0x8a, b0x8a, r0x8b, g0x8b, b0x8b;
    __m256i r1x8a, g1x8a, b1x8a, r1x8b, g1x8b, b1x8b;

    __m256i r0ox16, g0ox16, b0ox16;
    __m256i y0ox16;
    __m256i roax8, robx8, goax8, gobx8, boax8, bobx8;
    __m256i yoax8, yobx8;

    __m256i r1ox16, g1ox16, b1ox16;
    __m256i y1ox16;
    __m256i r1oax8, r1obx8, g1oax8, g1obx8, b1oax8, b1obx8;
    __m256i y1oax8, y1obx8, uvoax8, uvobx8, uvox16;
    __m256i uox8, vox8, ravgx8, gavgx8, bavgx8;
    for (; height > 1; height -= 2,
                       dsty += dstlinesize[0] * 2, dstuv += dstlinesize[1],
                       srcy += srclinesize[0], srcuv += srclinesize[1] / 2) {
        for (int xx = 0; xx < width >> 4; xx++) {
            int x = xx << 4;

            y0x16 = _mm256_lddqu_si256((__m256i*)(srcy + x));
            y1x16 = _mm256_lddqu_si256((__m256i*)(srcy + (srclinesize[0] / 2 + x)));
            uvx16 = _mm256_lddqu_si256((__m256i*)(srcuv + x));

            if (in_depth == 10) {
                // shift to low10bits for 10bit input
                y0x16 = _mm256_srli_epi16(y0x16, 6);
                y1x16 = _mm256_srli_epi16(y1x16, 6);
                uvx16 = _mm256_srli_epi16(uvx16, 6);
            }

            y0x8a = _mm256_cvtepi16_epi32(_mm256_extracti128_si256(y0x16, 0));
            y0x8b = _mm256_cvtepi16_epi32(_mm256_extracti128_si256(y0x16, 1));
            y1x8a = _mm256_cvtepi16_epi32(_mm256_extracti128_si256(y1x16, 0));
            y1x8b = _mm256_cvtepi16_epi32(_mm256_extracti128_si256(y1x16, 1));
            uvx8a = _mm256_cvtepi16_epi32(_mm256_extracti128_si256(uvx16, 0));
            uvx8b = _mm256_cvtepi16_epi32(_mm256_extracti128_si256(uvx16, 1));
            y0x8a = _mm256_sub_epi32(y0x8a, in_yuv_offx8);
            y1x8a = _mm256_sub_epi32(y1x8a, in_yuv_offx8);
            y0x8b = _mm256_sub_epi32(y0x8b, in_yuv_offx8);
            y1x8b = _mm256_sub_epi32(y1x8b, in_yuv_offx8);
            uvx8a = _mm256_sub_epi32(uvx8a, in_uv_offx8);
            uvx8b = _mm256_sub_epi32(uvx8b, in_uv_offx8);

            ux8a = _mm256_shuffle_epi32(uvx8a, _MM_SHUFFLE(2, 2, 0, 0));
            ux8b = _mm256_shuffle_epi32(uvx8b, _MM_SHUFFLE(2, 2, 0, 0));
            vx8a = _mm256_shuffle_epi32(uvx8a, _MM_SHUFFLE(3, 3, 1, 1));
            vx8b = _mm256_shuffle_epi32(uvx8b, _MM_SHUFFLE(3, 3, 1, 1));

            // r = av_clip_int16((y * cy + crv * v + in_rnd) >> in_sh);
            r0x8a = g0x8a = b0x8a = _mm256_mullo_epi32(y0x8a, cyx8);
            r0x8a = _mm256_add_epi32(r0x8a, _mm256_mullo_epi32(vx8a, _mm256_set1_epi32(crv)));
            r0x8a = _mm256_add_epi32(r0x8a, rndx8);
            r0x8a = _mm256_srai_epi32(r0x8a, in_sh);
            r0x8a = av_clip_int16_avx(r0x8a);

            r1x8a = g1x8a = b1x8a = _mm256_mullo_epi32(y1x8a, cyx8);
            r1x8a = _mm256_add_epi32(r1x8a, _mm256_mullo_epi32(vx8a, _mm256_set1_epi32(crv)));
            r1x8a = _mm256_add_epi32(r1x8a, rndx8);
            r1x8a = _mm256_srai_epi32(r1x8a, in_sh);
            r1x8a = av_clip_int16_avx(r1x8a);

            // g = av_clip_int16((y * cy + cgu * u + cgv * v + in_rnd) >> in_sh);
            g0x8a = _mm256_add_epi32(g0x8a, _mm256_mullo_epi32(ux8a, _mm256_set1_epi32(cgu)));
            g0x8a = _mm256_add_epi32(g0x8a, _mm256_mullo_epi32(vx8a, _mm256_set1_epi32(cgv)));
            g0x8a = _mm256_add_epi32(g0x8a, rndx8);
            g0x8a = _mm256_srai_epi32(g0x8a, in_sh);
            g0x8a = av_clip_int16_avx(g0x8a);

            g1x8a = _mm256_add_epi32(g1x8a, _mm256_mullo_epi32(ux8a, _mm256_set1_epi32(cgu)));
            g1x8a = _mm256_add_epi32(g1x8a, _mm256_mullo_epi32(vx8a, _mm256_set1_epi32(cgv)));
            g1x8a = _mm256_add_epi32(g1x8a, rndx8);
            g1x8a = _mm256_srai_epi32(g1x8a, in_sh);
            g1x8a = av_clip_int16_avx(g1x8a);

            // b = av_clip_int16((y * cy + cbu * u + in_rnd) >> in_sh);
            b0x8a = _mm256_add_epi32(b0x8a, _mm256_mullo_epi32(ux8a, _mm256_set1_epi32(cbu)));
            b0x8a = _mm256_add_epi32(b0x8a, rndx8);
            b0x8a = _mm256_srai_epi32(b0x8a, in_sh);
            b0x8a = av_clip_int16_avx(b0x8a);

            b1x8a = _mm256_add_epi32(b1x8a, _mm256_mullo_epi32(ux8a, _mm256_set1_epi32(cbu)));
            b1x8a = _mm256_add_epi32(b1x8a, rndx8);
            b1x8a = _mm256_srai_epi32(b1x8a, in_sh);
            b1x8a = av_clip_int16_avx(b1x8a);

            r0x8b = g0x8b = b0x8b = _mm256_mullo_epi32(y0x8b, cyx8);
            r0x8b = _mm256_add_epi32(r0x8b, _mm256_mullo_epi32(vx8b, _mm256_set1_epi32(crv)));
            r0x8b = _mm256_add_epi32(r0x8b, rndx8);
            r0x8b = _mm256_srai_epi32(r0x8b, in_sh);
            r0x8b = av_clip_int16_avx(r0x8b);

            r1x8b = g1x8b = b1x8b = _mm256_mullo_epi32(y1x8b, cyx8);
            r1x8b = _mm256_add_epi32(r1x8b, _mm256_mullo_epi32(vx8b, _mm256_set1_epi32(crv)));
            r1x8b = _mm256_add_epi32(r1x8b, rndx8);
            r1x8b = _mm256_srai_epi32(r1x8b, in_sh);
            r1x8b = av_clip_int16_avx(r1x8b);

            g0x8b = _mm256_add_epi32(g0x8b, _mm256_mullo_epi32(ux8b, _mm256_set1_epi32(cgu)));
            g0x8b = _mm256_add_epi32(g0x8b, _mm256_mullo_epi32(vx8b, _mm256_set1_epi32(cgv)));
            g0x8b = _mm256_add_epi32(g0x8b, rndx8);
            g0x8b = _mm256_srai_epi32(g0x8b, in_sh);
            g0x8b = av_clip_int16_avx(g0x8b);

            g1x8b = _mm256_add_epi32(g1x8b, _mm256_mullo_epi32(ux8b, _mm256_set1_epi32(cgu)));
            g1x8b = _mm256_add_epi32(g1x8b, _mm256_mullo_epi32(vx8b, _mm256_set1_epi32(cgv)));
            g1x8b = _mm256_add_epi32(g1x8b, rndx8);
            g1x8b = _mm256_srai_epi32(g1x8b, in_sh);
            g1x8b = av_clip_int16_avx(g1x8b);

            b0x8b = _mm256_add_epi32(b0x8b, _mm256_mullo_epi32(ux8b, _mm256_set1_epi32(cbu)));
            b0x8b = _mm256_add_epi32(b0x8b, rndx8);
            b0x8b = _mm256_srai_epi32(b0x8b, in_sh);
            b0x8b = av_clip_int16_avx(b0x8b);

            b1x8b = _mm256_add_epi32(b1x8b, _mm256_mullo_epi32(ux8b, _mm256_set1_epi32(cbu)));
            b1x8b = _mm256_add_epi32(b1x8b, rndx8);
            b1x8b = _mm256_srai_epi32(b1x8b, in_sh);
            b1x8b = av_clip_int16_avx(b1x8b);

            tonemap_int32x8_avx(r0x8a, g0x8a, b0x8a, r, g, b,
                                params->lin_lut, params->tonemap_lut, params->delin_lut,
                                params->coeffs, params->ocoeffs, params->desat, params->rgb2rgb_coeffs,
                                params->rgb2rgb_passthrough);
            tonemap_int32x8_avx(r1x8a, g1x8a, b1x8a, r1, g1, b1,
                                params->lin_lut, params->tonemap_lut, params->delin_lut,
                                params->coeffs, params->ocoeffs, params->desat, params->rgb2rgb_coeffs,
                                params->rgb2rgb_passthrough);
            tonemap_int32x8_avx(r0x8b, g0x8b, b0x8b, &r[8], &g[8], &b[8],
                                params->lin_lut, params->tonemap_lut, params->delin_lut,
                                params->coeffs, params->ocoeffs, params->desat, params->rgb2rgb_coeffs,
                                params->rgb2rgb_passthrough);
            tonemap_int32x8_avx(r1x8b, g1x8b, b1x8b, &r1[8], &g1[8], &b1[8],
                                params->lin_lut, params->tonemap_lut, params->delin_lut,
                                params->coeffs, params->ocoeffs, params->desat, params->rgb2rgb_coeffs,
                                params->rgb2rgb_passthrough);

            r0ox16 = _mm256_lddqu_si256((const __m256i_u *)r);
            g0ox16 = _mm256_lddqu_si256((const __m256i_u *)g);
            b0ox16 = _mm256_lddqu_si256((const __m256i_u *)b);

            roax8 = _mm256_cvtepi16_epi32(_mm256_extracti128_si256(r0ox16, 0));
            goax8 = _mm256_cvtepi16_epi32(_mm256_extracti128_si256(g0ox16, 0));
            boax8 = _mm256_cvtepi16_epi32(_mm256_extracti128_si256(b0ox16, 0));

            robx8 = _mm256_cvtepi16_epi32(_mm256_extracti128_si256(r0ox16, 1));
            gobx8 = _mm256_cvtepi16_epi32(_mm256_extracti128_si256(g0ox16, 1));
            bobx8 = _mm256_cvtepi16_epi32(_mm256_extracti128_si256(b0ox16, 1));

            yoax8 = _mm256_mullo_epi32(roax8, _mm256_set1_epi32(cry));
            yoax8 = _mm256_add_epi32(yoax8, _mm256_mullo_epi32(goax8, _mm256_set1_epi32(cgy)));
            yoax8 = _mm256_add_epi32(yoax8, _mm256_mullo_epi32(boax8, _mm256_set1_epi32(cby)));
            yoax8 = _mm256_add_epi32(yoax8, _mm256_set1_epi32(out_rnd));
            yoax8 = _mm256_srai_epi32(yoax8, out_sh);
            yoax8 = _mm256_add_epi32(yoax8, _mm256_set1_epi32(params->out_yuv_off));

            yobx8 = _mm256_mullo_epi32(robx8, _mm256_set1_epi32(cry));
            yobx8 = _mm256_add_epi32(yobx8, _mm256_mullo_epi32(gobx8, _mm256_set1_epi32(cgy)));
            yobx8 = _mm256_add_epi32(yobx8, _mm256_mullo_epi32(bobx8, _mm256_set1_epi32(cby)));
            yobx8 = _mm256_add_epi32(yobx8, _mm256_set1_epi32(out_rnd));
            yobx8 = _mm256_srai_epi32(yobx8, out_sh);
            yobx8 = _mm256_add_epi32(yobx8, _mm256_set1_epi32(params->out_yuv_off));

            y0ox16 = _mm256_packs_epi32(yoax8, yobx8);
            y0ox16 = _mm256_permute4x64_epi64(y0ox16, _MM_SHUFFLE(3, 1, 2, 0));
            _mm_storeu_si128((__m128i_u *) &dsty[x], _mm256_castsi256_si128(_mm256_permute4x64_epi64(_mm256_packus_epi16(y0ox16, _mm256_setzero_si256()), _MM_SHUFFLE(3, 1, 2, 0))));

            r1ox16 = _mm256_lddqu_si256((const __m256i_u *)r1);
            g1ox16 = _mm256_lddqu_si256((const __m256i_u *)g1);
            b1ox16 = _mm256_lddqu_si256((const __m256i_u *)b1);

            r1oax8 = _mm256_cvtepi16_epi32(_mm256_extracti128_si256(r1ox16, 0));
            g1oax8 = _mm256_cvtepi16_epi32(_mm256_extracti128_si256(g1ox16, 0));
            b1oax8 = _mm256_cvtepi16_epi32(_mm256_extracti128_si256(b1ox16, 0));

            r1obx8 = _mm256_cvtepi16_epi32(_mm256_extracti128_si256(r1ox16, 1));
            g1obx8 = _mm256_cvtepi16_epi32(_mm256_extracti128_si256(g1ox16, 1));
            b1obx8 = _mm256_cvtepi16_epi32(_mm256_extracti128_si256(b1ox16, 1));

            y1oax8 = _mm256_mullo_epi32(r1oax8, _mm256_set1_epi32(cry));
            y1oax8 = _mm256_add_epi32(y1oax8, _mm256_mullo_epi32(g1oax8, _mm256_set1_epi32(cgy)));
            y1oax8 = _mm256_add_epi32(y1oax8, _mm256_mullo_epi32(b1oax8, _mm256_set1_epi32(cby)));
            y1oax8 = _mm256_add_epi32(y1oax8, _mm256_set1_epi32(out_rnd));
            y1oax8 = _mm256_srai_epi32(y1oax8, out_sh);
            y1oax8 = _mm256_add_epi32(y1oax8, _mm256_set1_epi32(params->out_yuv_off));

            y1obx8 = _mm256_mullo_epi32(r1obx8, _mm256_set1_epi32(cry));
            y1obx8 = _mm256_add_epi32(y1obx8, _mm256_mullo_epi32(g1obx8, _mm256_set1_epi32(cgy)));
            y1obx8 = _mm256_add_epi32(y1obx8, _mm256_mullo_epi32(b1obx8, _mm256_set1_epi32(cby)));
            y1obx8 = _mm256_add_epi32(y1obx8, _mm256_set1_epi32(out_rnd));
            y1obx8 = _mm256_srai_epi32(y1obx8, out_sh);
            y1obx8 = _mm256_add_epi32(y1obx8, _mm256_set1_epi32(params->out_yuv_off));

            y1ox16 = _mm256_packs_epi32(y1oax8, y1obx8);
            y1ox16 = _mm256_permute4x64_epi64(y1ox16, _MM_SHUFFLE(3, 1, 2, 0));
            _mm_storeu_si128((__m128i_u *) &dsty[x + dstlinesize[0]], _mm256_castsi256_si128(_mm256_permute4x64_epi64(_mm256_packus_epi16(y1ox16, _mm256_setzero_si256()), _MM_SHUFFLE(3, 1, 2, 0))));

            ravgx8 = _mm256_hadd_epi32(roax8, robx8);
            ravgx8 = _mm256_add_epi32(ravgx8, _mm256_hadd_epi32(r1oax8, r1obx8));
            ravgx8 = _mm256_permute4x64_epi64(ravgx8, _MM_SHUFFLE(3, 1, 2, 0));
            ravgx8 = _mm256_add_epi32(ravgx8, _mm256_set1_epi32(2));
            ravgx8 = _mm256_srai_epi32(ravgx8, 2);

            gavgx8 = _mm256_hadd_epi32(goax8, gobx8);
            gavgx8 = _mm256_add_epi32(gavgx8, _mm256_hadd_epi32(g1oax8, g1obx8));
            gavgx8 = _mm256_permute4x64_epi64(gavgx8, _MM_SHUFFLE(3, 1, 2, 0));
            gavgx8 = _mm256_add_epi32(gavgx8, _mm256_set1_epi32(2));
            gavgx8 = _mm256_srai_epi32(gavgx8, 2);

            bavgx8 = _mm256_hadd_epi32(boax8, bobx8);
            bavgx8 = _mm256_add_epi32(bavgx8, _mm256_hadd_epi32(b1oax8, b1obx8));
            bavgx8 = _mm256_permute4x64_epi64(bavgx8, _MM_SHUFFLE(3, 1, 2, 0));
            bavgx8 = _mm256_add_epi32(bavgx8, _mm256_set1_epi32(2));
            bavgx8 = _mm256_srai_epi32(bavgx8, 2);

            uox8 = _mm256_add_epi32(_mm256_set1_epi32(out_rnd), _mm256_mullo_epi32(ravgx8, _mm256_set1_epi32(cru)));
            uox8 = _mm256_add_epi32(uox8, _mm256_mullo_epi32(gavgx8, _mm256_set1_epi32(ocgu)));
            uox8 = _mm256_add_epi32(uox8, _mm256_mullo_epi32(bavgx8, _mm256_set1_epi32(cburv)));
            uox8 = _mm256_srai_epi32(uox8, out_sh);
            uox8 = _mm256_add_epi32(uox8, _mm256_set1_epi32(out_uv_offset));

            vox8 = _mm256_add_epi32(_mm256_set1_epi32(out_rnd), _mm256_mullo_epi32(ravgx8, _mm256_set1_epi32(cburv)));
            vox8 = _mm256_add_epi32(vox8, _mm256_mullo_epi32(gavgx8, _mm256_set1_epi32(ocgv)));
            vox8 = _mm256_add_epi32(vox8, _mm256_mullo_epi32(bavgx8, _mm256_set1_epi32(cbv)));
            vox8 = _mm256_srai_epi32(vox8, out_sh);
            vox8 = _mm256_add_epi32(vox8, _mm256_set1_epi32(out_uv_offset));

            uvoax8 = _mm256_unpacklo_epi32(uox8, vox8);
            uvobx8 = _mm256_unpackhi_epi32(uox8, vox8);
            uvox16 = _mm256_packs_epi32(uvoax8, uvobx8);
            _mm_storeu_si128((__m128i_u *) &dstuv[x], _mm256_castsi256_si128(_mm256_permute4x64_epi64(_mm256_packus_epi16(uvox16, _mm256_setzero_si256()), _MM_SHUFFLE(3, 1, 2, 0))));
        }
    }

    // Process remaining pixels cannot fill the full simd register with scalar version
    if (remainw) {
        int offset = width & (int)0xfffffff0;
        rdsty += offset;
        rdstuv += offset;
        rsrcy += offset;
        rsrcuv += offset;
        tonemap_frame_p016_p010_2_nv12(rdsty, rdstuv,
                                       rsrcy, rsrcuv,
                                       dstlinesize, srclinesize,
                                       dstdepth, srcdepth,
                                       remainw, rheight, params);
    }
#endif // ENABLE_TONEMAPX_AVX_INTRINSICS
}

X86_64_V3 void tonemap_frame_p016_p010_2_p016_p010_avx(uint16_t *dsty, uint16_t *dstuv,
                                                       const uint16_t *srcy, const uint16_t *srcuv,
                                                       const int *dstlinesize, const int *srclinesize,
                                                       int dstdepth, int srcdepth,
                                                       int width, int height,
                                                       const struct TonemapIntParams *params)
{
#ifdef ENABLE_TONEMAPX_AVX_INTRINSICS
    uint16_t *rdsty = dsty;
    uint16_t *rdstuv = dstuv;
    const uint16_t *rsrcy = srcy;
    const uint16_t *rsrcuv = srcuv;
    int rheight = height;
    // not zero when not divisible by 8
    // intentionally leave last pixel emtpy when input is odd
    int remainw = width & 14;

    const int in_depth = srcdepth;
    const int in_uv_offset = 128 << (in_depth - 8);
    const int in_sh = in_depth - 1;
    const int in_rnd = 1 << (in_sh - 1);

    const int out_depth = dstdepth;
    const int out_uv_offset = 128 << (out_depth - 8);
    const int out_sh = 29 - out_depth;
    const int out_rnd = 1 << (out_sh - 1);
    const int out_sh2 = 16 - out_depth;

    int cy  = (*params->yuv2rgb_coeffs)[0][0][0];
    int crv = (*params->yuv2rgb_coeffs)[0][2][0];
    int cgu = (*params->yuv2rgb_coeffs)[1][1][0];
    int cgv = (*params->yuv2rgb_coeffs)[1][2][0];
    int cbu = (*params->yuv2rgb_coeffs)[2][1][0];

    int cry   = (*params->rgb2yuv_coeffs)[0][0][0];
    int cgy   = (*params->rgb2yuv_coeffs)[0][1][0];
    int cby   = (*params->rgb2yuv_coeffs)[0][2][0];
    int cru   = (*params->rgb2yuv_coeffs)[1][0][0];
    int ocgu  = (*params->rgb2yuv_coeffs)[1][1][0];
    int cburv = (*params->rgb2yuv_coeffs)[1][2][0];
    int ocgv  = (*params->rgb2yuv_coeffs)[2][1][0];
    int cbv   = (*params->rgb2yuv_coeffs)[2][2][0];

    int16_t r[16], g[16], b[16];
    int16_t r1[16], g1[16], b1[16];
    __m256i in_yuv_offx8 = _mm256_set1_epi32(params->in_yuv_off);
    __m256i in_uv_offx8 = _mm256_set1_epi32(in_uv_offset);
    __m256i cyx8 = _mm256_set1_epi32(cy);
    __m256i rndx8 = _mm256_set1_epi32(in_rnd);

    __m256i r0ox16, g0ox16, b0ox16;
    __m256i y0ox16;
    __m256i roax8, robx8, goax8, gobx8, boax8, bobx8;
    __m256i yoax8, yobx8;
    __m256i uvx16, uvx8a, uvx8b;
    __m256i y0x16, y1x16;
    __m256i y0x8a, y0x8b, y1x8a, y1x8b, ux8a, ux8b, vx8a, vx8b;
    __m256i r0x8a, g0x8a, b0x8a, r0x8b, g0x8b, b0x8b;
    __m256i r1x8a, g1x8a, b1x8a, r1x8b, g1x8b, b1x8b;

    __m256i r1ox16, g1ox16, b1ox16;
    __m256i y1ox16;
    __m256i r1oax8, r1obx8, g1oax8, g1obx8, b1oax8, b1obx8;
    __m256i y1oax8, y1obx8, uvoax8, uvobx8, uvox16;
    __m256i uox8, vox8, ravgx8, gavgx8, bavgx8;
    for (; height > 1; height -= 2,
                       dsty += dstlinesize[0], dstuv += dstlinesize[1] / 2,
                       srcy += srclinesize[0], srcuv += srclinesize[1] / 2) {
        for (int xx = 0; xx < width >> 4; xx++) {
            int x = xx << 4;

            y0x16 = _mm256_lddqu_si256((__m256i*)(srcy + x));
            y1x16 = _mm256_lddqu_si256((__m256i*)(srcy + (srclinesize[0] / 2 + x)));
            uvx16 = _mm256_lddqu_si256((__m256i*)(srcuv + x));

            if (in_depth == 10) {
                // shift to low10bits for 10bit input
                y0x16 = _mm256_srli_epi16(y0x16, 6);
                y1x16 = _mm256_srli_epi16(y1x16, 6);
                uvx16 = _mm256_srli_epi16(uvx16, 6);
            }

            y0x8a = _mm256_cvtepi16_epi32(_mm256_extracti128_si256(y0x16, 0));
            y0x8b = _mm256_cvtepi16_epi32(_mm256_extracti128_si256(y0x16, 1));
            y1x8a = _mm256_cvtepi16_epi32(_mm256_extracti128_si256(y1x16, 0));
            y1x8b = _mm256_cvtepi16_epi32(_mm256_extracti128_si256(y1x16, 1));
            uvx8a = _mm256_cvtepi16_epi32(_mm256_extracti128_si256(uvx16, 0));
            uvx8b = _mm256_cvtepi16_epi32(_mm256_extracti128_si256(uvx16, 1));
            y0x8a = _mm256_sub_epi32(y0x8a, in_yuv_offx8);
            y1x8a = _mm256_sub_epi32(y1x8a, in_yuv_offx8);
            y0x8b = _mm256_sub_epi32(y0x8b, in_yuv_offx8);
            y1x8b = _mm256_sub_epi32(y1x8b, in_yuv_offx8);
            uvx8a = _mm256_sub_epi32(uvx8a, in_uv_offx8);
            uvx8b = _mm256_sub_epi32(uvx8b, in_uv_offx8);

            ux8a = _mm256_shuffle_epi32(uvx8a, _MM_SHUFFLE(2, 2, 0, 0));
            ux8b = _mm256_shuffle_epi32(uvx8b, _MM_SHUFFLE(2, 2, 0, 0));
            vx8a = _mm256_shuffle_epi32(uvx8a, _MM_SHUFFLE(3, 3, 1, 1));
            vx8b = _mm256_shuffle_epi32(uvx8b, _MM_SHUFFLE(3, 3, 1, 1));

            // r = av_clip_int16((y * cy + crv * v + in_rnd) >> in_sh);
            r0x8a = g0x8a = b0x8a = _mm256_mullo_epi32(y0x8a, cyx8);
            r0x8a = _mm256_add_epi32(r0x8a, _mm256_mullo_epi32(vx8a, _mm256_set1_epi32(crv)));
            r0x8a = _mm256_add_epi32(r0x8a, rndx8);
            r0x8a = _mm256_srai_epi32(r0x8a, in_sh);
            r0x8a = av_clip_int16_avx(r0x8a);

            r1x8a = g1x8a = b1x8a = _mm256_mullo_epi32(y1x8a, cyx8);
            r1x8a = _mm256_add_epi32(r1x8a, _mm256_mullo_epi32(vx8a, _mm256_set1_epi32(crv)));
            r1x8a = _mm256_add_epi32(r1x8a, rndx8);
            r1x8a = _mm256_srai_epi32(r1x8a, in_sh);
            r1x8a = av_clip_int16_avx(r1x8a);

            // g = av_clip_int16((y * cy + cgu * u + cgv * v + in_rnd) >> in_sh);
            g0x8a = _mm256_add_epi32(g0x8a, _mm256_mullo_epi32(ux8a, _mm256_set1_epi32(cgu)));
            g0x8a = _mm256_add_epi32(g0x8a, _mm256_mullo_epi32(vx8a, _mm256_set1_epi32(cgv)));
            g0x8a = _mm256_add_epi32(g0x8a, rndx8);
            g0x8a = _mm256_srai_epi32(g0x8a, in_sh);
            g0x8a = av_clip_int16_avx(g0x8a);

            g1x8a = _mm256_add_epi32(g1x8a, _mm256_mullo_epi32(ux8a, _mm256_set1_epi32(cgu)));
            g1x8a = _mm256_add_epi32(g1x8a, _mm256_mullo_epi32(vx8a, _mm256_set1_epi32(cgv)));
            g1x8a = _mm256_add_epi32(g1x8a, rndx8);
            g1x8a = _mm256_srai_epi32(g1x8a, in_sh);
            g1x8a = av_clip_int16_avx(g1x8a);

            // b = av_clip_int16((y * cy + cbu * u + in_rnd) >> in_sh);
            b0x8a = _mm256_add_epi32(b0x8a, _mm256_mullo_epi32(ux8a, _mm256_set1_epi32(cbu)));
            b0x8a = _mm256_add_epi32(b0x8a, rndx8);
            b0x8a = _mm256_srai_epi32(b0x8a, in_sh);
            b0x8a = av_clip_int16_avx(b0x8a);

            b1x8a = _mm256_add_epi32(b1x8a, _mm256_mullo_epi32(ux8a, _mm256_set1_epi32(cbu)));
            b1x8a = _mm256_add_epi32(b1x8a, rndx8);
            b1x8a = _mm256_srai_epi32(b1x8a, in_sh);
            b1x8a = av_clip_int16_avx(b1x8a);

            r0x8b = g0x8b = b0x8b = _mm256_mullo_epi32(y0x8b, cyx8);
            r0x8b = _mm256_add_epi32(r0x8b, _mm256_mullo_epi32(vx8b, _mm256_set1_epi32(crv)));
            r0x8b = _mm256_add_epi32(r0x8b, rndx8);
            r0x8b = _mm256_srai_epi32(r0x8b, in_sh);
            r0x8b = av_clip_int16_avx(r0x8b);

            r1x8b = g1x8b = b1x8b = _mm256_mullo_epi32(y1x8b, cyx8);
            r1x8b = _mm256_add_epi32(r1x8b, _mm256_mullo_epi32(vx8b, _mm256_set1_epi32(crv)));
            r1x8b = _mm256_add_epi32(r1x8b, rndx8);
            r1x8b = _mm256_srai_epi32(r1x8b, in_sh);
            r1x8b = av_clip_int16_avx(r1x8b);

            g0x8b = _mm256_add_epi32(g0x8b, _mm256_mullo_epi32(ux8b, _mm256_set1_epi32(cgu)));
            g0x8b = _mm256_add_epi32(g0x8b, _mm256_mullo_epi32(vx8b, _mm256_set1_epi32(cgv)));
            g0x8b = _mm256_add_epi32(g0x8b, rndx8);
            g0x8b = _mm256_srai_epi32(g0x8b, in_sh);
            g0x8b = av_clip_int16_avx(g0x8b);

            g1x8b = _mm256_add_epi32(g1x8b, _mm256_mullo_epi32(ux8b, _mm256_set1_epi32(cgu)));
            g1x8b = _mm256_add_epi32(g1x8b, _mm256_mullo_epi32(vx8b, _mm256_set1_epi32(cgv)));
            g1x8b = _mm256_add_epi32(g1x8b, rndx8);
            g1x8b = _mm256_srai_epi32(g1x8b, in_sh);
            g1x8b = av_clip_int16_avx(g1x8b);

            b0x8b = _mm256_add_epi32(b0x8b, _mm256_mullo_epi32(ux8b, _mm256_set1_epi32(cbu)));
            b0x8b = _mm256_add_epi32(b0x8b, rndx8);
            b0x8b = _mm256_srai_epi32(b0x8b, in_sh);
            b0x8b = av_clip_int16_avx(b0x8b);

            b1x8b = _mm256_add_epi32(b1x8b, _mm256_mullo_epi32(ux8b, _mm256_set1_epi32(cbu)));
            b1x8b = _mm256_add_epi32(b1x8b, rndx8);
            b1x8b = _mm256_srai_epi32(b1x8b, in_sh);
            b1x8b = av_clip_int16_avx(b1x8b);

            tonemap_int32x8_avx(r0x8a, g0x8a, b0x8a, r, g, b,
                                params->lin_lut, params->tonemap_lut, params->delin_lut,
                                params->coeffs, params->ocoeffs, params->desat, params->rgb2rgb_coeffs,
                                params->rgb2rgb_passthrough);
            tonemap_int32x8_avx(r1x8a, g1x8a, b1x8a, r1, g1, b1,
                                params->lin_lut, params->tonemap_lut, params->delin_lut,
                                params->coeffs, params->ocoeffs, params->desat, params->rgb2rgb_coeffs,
                                params->rgb2rgb_passthrough);
            tonemap_int32x8_avx(r0x8b, g0x8b, b0x8b, &r[8], &g[8], &b[8],
                                params->lin_lut, params->tonemap_lut, params->delin_lut,
                                params->coeffs, params->ocoeffs, params->desat, params->rgb2rgb_coeffs,
                                params->rgb2rgb_passthrough);
            tonemap_int32x8_avx(r1x8b, g1x8b, b1x8b, &r1[8], &g1[8], &b1[8],
                                params->lin_lut, params->tonemap_lut, params->delin_lut,
                                params->coeffs, params->ocoeffs, params->desat, params->rgb2rgb_coeffs,
                                params->rgb2rgb_passthrough);

            r0ox16 = _mm256_lddqu_si256((const __m256i_u *)r);
            g0ox16 = _mm256_lddqu_si256((const __m256i_u *)g);
            b0ox16 = _mm256_lddqu_si256((const __m256i_u *)b);

            roax8 = _mm256_cvtepi16_epi32(_mm256_extracti128_si256(r0ox16, 0));
            goax8 = _mm256_cvtepi16_epi32(_mm256_extracti128_si256(g0ox16, 0));
            boax8 = _mm256_cvtepi16_epi32(_mm256_extracti128_si256(b0ox16, 0));

            robx8 = _mm256_cvtepi16_epi32(_mm256_extracti128_si256(r0ox16, 1));
            gobx8 = _mm256_cvtepi16_epi32(_mm256_extracti128_si256(g0ox16, 1));
            bobx8 = _mm256_cvtepi16_epi32(_mm256_extracti128_si256(b0ox16, 1));

            yoax8 = _mm256_mullo_epi32(roax8, _mm256_set1_epi32(cry));
            yoax8 = _mm256_add_epi32(yoax8, _mm256_mullo_epi32(goax8, _mm256_set1_epi32(cgy)));
            yoax8 = _mm256_add_epi32(yoax8, _mm256_mullo_epi32(boax8, _mm256_set1_epi32(cby)));
            yoax8 = _mm256_add_epi32(yoax8, _mm256_set1_epi32(out_rnd));
            yoax8 = _mm256_srai_epi32(yoax8, out_sh);
            yoax8 = _mm256_add_epi32(yoax8, _mm256_set1_epi32(params->out_yuv_off));

            yobx8 = _mm256_mullo_epi32(robx8, _mm256_set1_epi32(cry));
            yobx8 = _mm256_add_epi32(yobx8, _mm256_mullo_epi32(gobx8, _mm256_set1_epi32(cgy)));
            yobx8 = _mm256_add_epi32(yobx8, _mm256_mullo_epi32(bobx8, _mm256_set1_epi32(cby)));
            yobx8 = _mm256_add_epi32(yobx8, _mm256_set1_epi32(out_rnd));
            yobx8 = _mm256_srai_epi32(yobx8, out_sh);
            yobx8 = _mm256_add_epi32(yobx8, _mm256_set1_epi32(params->out_yuv_off));

            y0ox16 = _mm256_packus_epi32(yoax8, yobx8);
            y0ox16 = _mm256_permute4x64_epi64(y0ox16, _MM_SHUFFLE(3, 1, 2, 0));
            y0ox16 = _mm256_slli_epi16(y0ox16, out_sh2);
            _mm256_storeu_si256((__m256i_u *) &dsty[x], y0ox16);

            r1ox16 = _mm256_lddqu_si256((const __m256i_u *)r1);
            g1ox16 = _mm256_lddqu_si256((const __m256i_u *)g1);
            b1ox16 = _mm256_lddqu_si256((const __m256i_u *)b1);

            r1oax8 = _mm256_cvtepi16_epi32(_mm256_extracti128_si256(r1ox16, 0));
            g1oax8 = _mm256_cvtepi16_epi32(_mm256_extracti128_si256(g1ox16, 0));
            b1oax8 = _mm256_cvtepi16_epi32(_mm256_extracti128_si256(b1ox16, 0));

            r1obx8 = _mm256_cvtepi16_epi32(_mm256_extracti128_si256(r1ox16, 1));
            g1obx8 = _mm256_cvtepi16_epi32(_mm256_extracti128_si256(g1ox16, 1));
            b1obx8 = _mm256_cvtepi16_epi32(_mm256_extracti128_si256(b1ox16, 1));

            y1oax8 = _mm256_mullo_epi32(r1oax8, _mm256_set1_epi32(cry));
            y1oax8 = _mm256_add_epi32(y1oax8, _mm256_mullo_epi32(g1oax8, _mm256_set1_epi32(cgy)));
            y1oax8 = _mm256_add_epi32(y1oax8, _mm256_mullo_epi32(b1oax8, _mm256_set1_epi32(cby)));
            y1oax8 = _mm256_add_epi32(y1oax8, _mm256_set1_epi32(out_rnd));
            y1oax8 = _mm256_srai_epi32(y1oax8, out_sh);
            y1oax8 = _mm256_add_epi32(y1oax8, _mm256_set1_epi32(params->out_yuv_off));

            y1obx8 = _mm256_mullo_epi32(r1obx8, _mm256_set1_epi32(cry));
            y1obx8 = _mm256_add_epi32(y1obx8, _mm256_mullo_epi32(g1obx8, _mm256_set1_epi32(cgy)));
            y1obx8 = _mm256_add_epi32(y1obx8, _mm256_mullo_epi32(b1obx8, _mm256_set1_epi32(cby)));
            y1obx8 = _mm256_add_epi32(y1obx8, _mm256_set1_epi32(out_rnd));
            y1obx8 = _mm256_srai_epi32(y1obx8, out_sh);
            y1obx8 = _mm256_add_epi32(y1obx8, _mm256_set1_epi32(params->out_yuv_off));

            y1ox16 = _mm256_packus_epi32(y1oax8, y1obx8);
            y1ox16 = _mm256_permute4x64_epi64(y1ox16, _MM_SHUFFLE(3, 1, 2, 0));
            y1ox16 = _mm256_slli_epi16(y1ox16, out_sh2);
            _mm256_storeu_si256((__m256i_u *) &dsty[x + dstlinesize[0] / 2], y1ox16);

            ravgx8 = _mm256_hadd_epi32(roax8, robx8);
            ravgx8 = _mm256_add_epi32(ravgx8, _mm256_hadd_epi32(r1oax8, r1obx8));
            ravgx8 = _mm256_permute4x64_epi64(ravgx8, _MM_SHUFFLE(3, 1, 2, 0));
            ravgx8 = _mm256_add_epi32(ravgx8, _mm256_set1_epi32(2));
            ravgx8 = _mm256_srai_epi32(ravgx8, 2);

            gavgx8 = _mm256_hadd_epi32(goax8, gobx8);
            gavgx8 = _mm256_add_epi32(gavgx8, _mm256_hadd_epi32(g1oax8, g1obx8));
            gavgx8 = _mm256_permute4x64_epi64(gavgx8, _MM_SHUFFLE(3, 1, 2, 0));
            gavgx8 = _mm256_add_epi32(gavgx8, _mm256_set1_epi32(2));
            gavgx8 = _mm256_srai_epi32(gavgx8, 2);

            bavgx8 = _mm256_hadd_epi32(boax8, bobx8);
            bavgx8 = _mm256_add_epi32(bavgx8, _mm256_hadd_epi32(b1oax8, b1obx8));
            bavgx8 = _mm256_permute4x64_epi64(bavgx8, _MM_SHUFFLE(3, 1, 2, 0));
            bavgx8 = _mm256_add_epi32(bavgx8, _mm256_set1_epi32(2));
            bavgx8 = _mm256_srai_epi32(bavgx8, 2);

            uox8 = _mm256_add_epi32(_mm256_set1_epi32(out_rnd), _mm256_mullo_epi32(ravgx8, _mm256_set1_epi32(cru)));
            uox8 = _mm256_add_epi32(uox8, _mm256_mullo_epi32(gavgx8, _mm256_set1_epi32(ocgu)));
            uox8 = _mm256_add_epi32(uox8, _mm256_mullo_epi32(bavgx8, _mm256_set1_epi32(cburv)));
            uox8 = _mm256_srai_epi32(uox8, out_sh);
            uox8 = _mm256_add_epi32(uox8, _mm256_set1_epi32(out_uv_offset));

            vox8 = _mm256_add_epi32(_mm256_set1_epi32(out_rnd), _mm256_mullo_epi32(ravgx8, _mm256_set1_epi32(cburv)));
            vox8 = _mm256_add_epi32(vox8, _mm256_mullo_epi32(gavgx8, _mm256_set1_epi32(ocgv)));
            vox8 = _mm256_add_epi32(vox8, _mm256_mullo_epi32(bavgx8, _mm256_set1_epi32(cbv)));
            vox8 = _mm256_srai_epi32(vox8, out_sh);
            vox8 = _mm256_add_epi32(vox8, _mm256_set1_epi32(out_uv_offset));

            uvoax8 = _mm256_unpacklo_epi32(uox8, vox8);
            uvobx8 = _mm256_unpackhi_epi32(uox8, vox8);
            uvox16 = _mm256_packus_epi32(uvoax8, uvobx8);
            uvox16 = _mm256_slli_epi16(uvox16, out_sh2);
            _mm256_storeu_si256((__m256i_u *) &dstuv[x], uvox16);
        }
    }

    // Process remaining pixels cannot fill the full simd register with scalar version
    if (remainw) {
        int offset = width & (int)0xfffffff0;
        rdsty += offset;
        rdstuv += offset;
        rsrcy += offset;
        rsrcuv += offset;
        tonemap_frame_p016_p010_2_p016_p010(rdsty, rdstuv,
                                            rsrcy, rsrcuv,
                                            dstlinesize, srclinesize,
                                            dstdepth, srcdepth,
                                            remainw, rheight, params);
    }
#endif // ENABLE_TONEMAPX_AVX_INTRINSICS
}
