/*
 * Copyright (c) 2024 Gnattu OC <gnattuoc@me.com>
 *
 * This file is part of FFmpeg.
 *
 * FFmpeg is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * FFmpeg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with FFmpeg; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

#include "vf_tonemapx_intrin_sse.h"

#ifdef ENABLE_TONEMAPX_SSE_INTRINSICS
#    include <immintrin.h>
#endif // ENABLE_TONEMAPX_SSE_INTRINSICS

#ifdef ENABLE_TONEMAPX_SSE_INTRINSICS
// GCC 10 and below does not implement _mm_storeu_si32 with movd instruction
// cast the register into float register and store with movss as a workaround
#if (defined(__GNUC__) && !defined(__clang__)) && (__GNUC__ <= 10)
__attribute__((always_inline))
X86_64_V2 static inline void _mm_storeu_si32(void* mem_addr, __m128i a) {
    _mm_store_ss((float*)mem_addr, _mm_castsi128_ps(a));
    return;
}
#endif

X86_64_V2 static inline __m128i av_clip_uint16_sse(__m128i a)
{
__m128i mask = _mm_set1_epi32(0x7FFF);
__m128i condition = _mm_and_si128(a, _mm_set1_epi32(~0x7FFF));

__m128i zero = _mm_setzero_si128();
__m128i cmp = _mm_cmpeq_epi32(condition, zero);

__m128i neg_a = _mm_and_si128(_mm_srai_epi32(_mm_xor_si128(a, _mm_set1_epi32(-1)), 31), mask);
__m128i result = _mm_or_si128(_mm_and_si128(cmp, a), _mm_andnot_si128(cmp, neg_a));

return result;
}

X86_64_V2 static inline __m128i av_clip_int16_sse(__m128i a)
{
__m128i add_result = _mm_add_epi32(a, _mm_set1_epi32(0x8000U));
__m128i mask = _mm_set1_epi32(~0xFFFF);
__m128i condition = _mm_and_si128(add_result, mask);
__m128i cmp = _mm_cmpeq_epi32(condition, _mm_setzero_si128());

__m128i shifted = _mm_srai_epi32(a, 31);
__m128i xor_result = _mm_xor_si128(shifted, _mm_set1_epi32(0x7FFF));

return _mm_or_si128(_mm_and_si128(cmp, a), _mm_andnot_si128(cmp, xor_result));
}

X86_64_V2 static inline void tonemap_int32x4_sse(__m128i r_in, __m128i g_in, __m128i b_in,
                                                 int16_t *r_out, int16_t *g_out, int16_t *b_out,
                                                 float *lin_lut, float *tonemap_lut, uint16_t *delin_lut,
                                                 const AVLumaCoefficients *coeffs,
                                                 const AVLumaCoefficients *ocoeffs, double desat,
                                                 double (*rgb2rgb)[3][3],
                                                 int rgb2rgb_passthrough)
{
    __m128i sig4;
    __m128 mapvalx4, r_linx4, g_linx4, b_linx4;
    __m128 offset = _mm_set1_ps(0.5f);
    __m128i input_lut_offset = _mm_set1_epi32(2048);
    __m128 intermediate_upper_bound = _mm_set1_ps(32767.0f);
    __m128i r, g, b, rx4, gx4, bx4;

    float mapval4[4], r_lin4[4], g_lin4[4], b_lin4[4];

    sig4 = _mm_max_epi32(r_in, _mm_max_epi32(g_in, b_in));
    sig4 = _mm_add_epi32(sig4, input_lut_offset);
    sig4 = av_clip_uint16_sse(sig4);

    r = _mm_add_epi32(r_in, input_lut_offset);
    r = av_clip_uint16_sse(r);
    g = _mm_add_epi32(g_in, input_lut_offset);
    g = av_clip_uint16_sse(g);
    b = _mm_add_epi32(b_in, input_lut_offset);
    b = av_clip_uint16_sse(b);

    // Cannot use loop here as the lane has to be compile-time constant
#define LOAD_LUT(i) mapval4[i] = tonemap_lut[_mm_extract_epi32(sig4, i)]; \
r_lin4[i] = lin_lut[_mm_extract_epi32(r, i)];                             \
g_lin4[i] = lin_lut[_mm_extract_epi32(g, i)];                             \
b_lin4[i] = lin_lut[_mm_extract_epi32(b, i)];

    LOAD_LUT(0)
    LOAD_LUT(1)
    LOAD_LUT(2)
    LOAD_LUT(3)

#undef LOAD_LUT

    mapvalx4 = _mm_loadu_ps(mapval4);
    r_linx4 = _mm_loadu_ps(r_lin4);
    g_linx4 = _mm_loadu_ps(g_lin4);
    b_linx4 = _mm_loadu_ps(b_lin4);

    if (!rgb2rgb_passthrough) {
        r_linx4 = _mm_mul_ps(r_linx4, _mm_set1_ps((float)(*rgb2rgb)[0][0]));
        r_linx4 = _mm_add_ps(r_linx4, _mm_mul_ps(g_linx4, _mm_set1_ps((float)(*rgb2rgb)[0][1])));
        r_linx4 = _mm_add_ps(r_linx4, _mm_mul_ps(b_linx4, _mm_set1_ps((float)(*rgb2rgb)[0][2])));

        g_linx4 = _mm_mul_ps(g_linx4, _mm_set1_ps((float)(*rgb2rgb)[1][1]));
        g_linx4 = _mm_add_ps(g_linx4, _mm_mul_ps(r_linx4, _mm_set1_ps((float)(*rgb2rgb)[1][0])));
        g_linx4 = _mm_add_ps(g_linx4, _mm_mul_ps(b_linx4, _mm_set1_ps((float)(*rgb2rgb)[1][2])));

        b_linx4 = _mm_mul_ps(b_linx4, _mm_set1_ps((float)(*rgb2rgb)[2][2]));
        b_linx4 = _mm_add_ps(b_linx4, _mm_mul_ps(r_linx4, _mm_set1_ps((float)(*rgb2rgb)[2][0])));
        b_linx4 = _mm_add_ps(b_linx4, _mm_mul_ps(g_linx4, _mm_set1_ps((float)(*rgb2rgb)[2][1])));
    }

    if (desat > 0) {
        __m128 eps_x4 = _mm_set1_ps(FLOAT_EPS);
        __m128 desat4 = _mm_set1_ps((float)desat);
        __m128 luma4 = _mm_set1_ps(0);
        __m128 overbright4;

        luma4 = _mm_add_ps(luma4, _mm_mul_ps(r_linx4, _mm_set1_ps((float)av_q2d(coeffs->cr))));
        luma4 = _mm_add_ps(luma4, _mm_mul_ps(g_linx4, _mm_set1_ps((float)av_q2d(coeffs->cg))));
        luma4 = _mm_add_ps(luma4, _mm_mul_ps(b_linx4, _mm_set1_ps((float)av_q2d(coeffs->cb))));
        overbright4 = _mm_div_ps(_mm_max_ps(_mm_sub_ps(luma4, desat4), eps_x4), _mm_max_ps(luma4, eps_x4));
        r_linx4 = _mm_sub_ps(r_linx4, _mm_mul_ps(r_linx4, overbright4));
        r_linx4 = _mm_add_ps(r_linx4, _mm_mul_ps(luma4, overbright4));
        g_linx4 = _mm_sub_ps(g_linx4, _mm_mul_ps(g_linx4, overbright4));
        g_linx4 = _mm_add_ps(g_linx4, _mm_mul_ps(luma4, overbright4));
        b_linx4 = _mm_sub_ps(b_linx4, _mm_mul_ps(b_linx4, overbright4));
        b_linx4 = _mm_add_ps(b_linx4, _mm_mul_ps(luma4, overbright4));
    }

    r_linx4 = _mm_mul_ps(r_linx4, mapvalx4);
    g_linx4 = _mm_mul_ps(g_linx4, mapvalx4);
    b_linx4 = _mm_mul_ps(b_linx4, mapvalx4);

    r_linx4 = _mm_mul_ps(r_linx4, intermediate_upper_bound);
    r_linx4 = _mm_add_ps(r_linx4, offset);

    g_linx4 = _mm_mul_ps(g_linx4, intermediate_upper_bound);
    g_linx4 = _mm_add_ps(g_linx4, offset);

    b_linx4 = _mm_mul_ps(b_linx4, intermediate_upper_bound);
    b_linx4 = _mm_add_ps(b_linx4, offset);

    rx4 = _mm_cvttps_epi32(r_linx4);
    rx4 = av_clip_uint16_sse(rx4);
    gx4 = _mm_cvttps_epi32(g_linx4);
    gx4 = av_clip_uint16_sse(gx4);
    bx4 = _mm_cvttps_epi32(b_linx4);
    bx4 = av_clip_uint16_sse(bx4);

#define SAVE_COLOR(i) r_out[i] = delin_lut[_mm_extract_epi32(rx4, i)]; \
g_out[i] = delin_lut[_mm_extract_epi32(gx4, i)];                       \
b_out[i] = delin_lut[_mm_extract_epi32(bx4, i)];

    SAVE_COLOR(0)
    SAVE_COLOR(1)
    SAVE_COLOR(2)
    SAVE_COLOR(3)

#undef SAVE_COLOR
}
#endif // ENABLE_TONEMAPX_SSE_INTRINSICS

X86_64_V2 void tonemap_frame_420p10_2_420p_sse(uint8_t *dsty, uint8_t *dstu, uint8_t *dstv,
                                               const uint16_t *srcy, const uint16_t *srcu, const uint16_t *srcv,
                                               const int *dstlinesize, const int *srclinesize,
                                               int dstdepth, int srcdepth,
                                               int width, int height,
                                               const struct TonemapIntParams *params)
{
#ifdef ENABLE_TONEMAPX_SSE_INTRINSICS
    uint8_t *rdsty = dsty;
    uint8_t *rdstu = dstu;
    uint8_t *rdstv = dstv;

    const uint16_t *rsrcy = srcy;
    const uint16_t *rsrcu = srcu;
    const uint16_t *rsrcv = srcv;

    int rheight = height;
    // not zero when not divisible by 8
    // intentionally leave last pixel emtpy when input is odd
    int remainw = width & 6;

    const int in_depth = srcdepth;
    const int in_uv_offset = 128 << (in_depth - 8);
    const int in_sh = in_depth - 1;
    const int in_rnd = 1 << (in_sh - 1);

    const int out_depth = dstdepth;
    const int out_uv_offset = 128 << (out_depth - 8);
    const int out_sh = 29 - out_depth;
    const int out_rnd = 1 << (out_sh - 1);

    int cy  = (*params->yuv2rgb_coeffs)[0][0][0];
    int crv = (*params->yuv2rgb_coeffs)[0][2][0];
    int cgu = (*params->yuv2rgb_coeffs)[1][1][0];
    int cgv = (*params->yuv2rgb_coeffs)[1][2][0];
    int cbu = (*params->yuv2rgb_coeffs)[2][1][0];

    int cry   = (*params->rgb2yuv_coeffs)[0][0][0];
    int cgy   = (*params->rgb2yuv_coeffs)[0][1][0];
    int cby   = (*params->rgb2yuv_coeffs)[0][2][0];
    int cru   = (*params->rgb2yuv_coeffs)[1][0][0];
    int ocgu  = (*params->rgb2yuv_coeffs)[1][1][0];
    int cburv = (*params->rgb2yuv_coeffs)[1][2][0];
    int ocgv  = (*params->rgb2yuv_coeffs)[2][1][0];
    int cbv   = (*params->rgb2yuv_coeffs)[2][2][0];

    int16_t r[8], g[8], b[8];
    int16_t r1[8], g1[8], b1[8];

    __m128i in_yuv_offx4 = _mm_set1_epi32(params->in_yuv_off);
    __m128i in_uv_offx4= _mm_set1_epi32(in_uv_offset);
    __m128i cyx4 = _mm_set1_epi32(cy);
    __m128i rndx4 = _mm_set1_epi32(in_rnd);
    __m128i zero128 = _mm_setzero_si128();
    __m128i ux4, vx4;
    __m128i y0x8, y1x8;
    __m128i y0x4a, y0x4b, y1x4a, y1x4b, ux4a, ux4b, vx4a, vx4b;
    __m128i r0x4a, g0x4a, b0x4a, r0x4b, g0x4b, b0x4b;
    __m128i r1x4a, g1x4a, b1x4a, r1x4b, g1x4b, b1x4b;

    __m128i r0ox8, g0ox8, b0ox8;
    __m128i y0ox8;
    __m128i roax4, robx4, goax4, gobx4, boax4, bobx4;
    __m128i yoax4, yobx4;

    __m128i r1ox8, g1ox8, b1ox8;
    __m128i y1ox8;
    __m128i r1oax4, r1obx4, g1oax4, g1obx4, b1oax4, b1obx4;
    __m128i y1oax4, y1obx4;
    __m128i uox4, vox4, ravgx4, gavgx4, bavgx4;
    for (; height > 1; height -= 2,
                       dsty += dstlinesize[0] * 2, dstu += dstlinesize[1], dstv += dstlinesize[2],
                       srcy += srclinesize[0], srcu += srclinesize[1] / 2, srcv += srclinesize[2] / 2) {
        for (int xx = 0; xx < width >> 3; xx++) {
            int x = xx << 3;

            y0x8 = _mm_lddqu_si128((__m128i*)(srcy + x));
            y1x8 = _mm_lddqu_si128((__m128i*)(srcy + (srclinesize[0] / 2 + x)));
            ux4 = _mm_loadu_si64((__m128i*)(srcu + (x >> 1)));
            vx4 = _mm_loadu_si64((__m128i*)(srcv + (x >> 1)));

            y0x4a = _mm_cvtepu16_epi32(y0x8);
            y0x4b = _mm_unpackhi_epi16(y0x8, zero128);
            y1x4a = _mm_cvtepu16_epi32(y1x8);
            y1x4b = _mm_unpackhi_epi16(y1x8, zero128);
            ux4 = _mm_cvtepu16_epi32(ux4);
            vx4 = _mm_cvtepu16_epi32(vx4);
            y0x4a = _mm_sub_epi32(y0x4a, in_yuv_offx4);
            y1x4a = _mm_sub_epi32(y1x4a, in_yuv_offx4);
            y0x4b = _mm_sub_epi32(y0x4b, in_yuv_offx4);
            y1x4b = _mm_sub_epi32(y1x4b, in_yuv_offx4);
            ux4 = _mm_sub_epi32(ux4, in_uv_offx4);
            vx4 = _mm_sub_epi32(vx4, in_uv_offx4);

            ux4a = _mm_unpacklo_epi32(ux4, ux4);
            ux4b = _mm_unpackhi_epi32(ux4, ux4);
            vx4a = _mm_unpacklo_epi32(vx4, vx4);
            vx4b = _mm_unpackhi_epi32(vx4, vx4);

            // r = av_clip_int16((y * cy + crv * v + in_rnd) >> in_sh);
            r0x4a = g0x4a = b0x4a = _mm_mullo_epi32(y0x4a, cyx4);
            r0x4a = _mm_add_epi32(r0x4a, _mm_mullo_epi32(vx4a, _mm_set1_epi32(crv)));
            r0x4a = _mm_add_epi32(r0x4a, rndx4);
            r0x4a = _mm_srai_epi32(r0x4a, in_sh);
            r0x4a = av_clip_int16_sse(r0x4a);

            r1x4a = g1x4a = b1x4a = _mm_mullo_epi32(y1x4a, cyx4);
            r1x4a = _mm_add_epi32(r1x4a, _mm_mullo_epi32(vx4a, _mm_set1_epi32(crv)));
            r1x4a = _mm_add_epi32(r1x4a, rndx4);
            r1x4a = _mm_srai_epi32(r1x4a, in_sh);
            r1x4a = av_clip_int16_sse(r1x4a);

            // g = av_clip_int16((y * cy + cgu * u + cgv * v + in_rnd) >> in_sh);
            g0x4a = _mm_add_epi32(g0x4a, _mm_mullo_epi32(ux4a, _mm_set1_epi32(cgu)));
            g0x4a = _mm_add_epi32(g0x4a, _mm_mullo_epi32(vx4a, _mm_set1_epi32(cgv)));
            g0x4a = _mm_add_epi32(g0x4a, rndx4);
            g0x4a = _mm_srai_epi32(g0x4a, in_sh);
            g0x4a = av_clip_int16_sse(g0x4a);

            g1x4a = _mm_add_epi32(g1x4a, _mm_mullo_epi32(ux4a, _mm_set1_epi32(cgu)));
            g1x4a = _mm_add_epi32(g1x4a, _mm_mullo_epi32(vx4a, _mm_set1_epi32(cgv)));
            g1x4a = _mm_add_epi32(g1x4a, rndx4);
            g1x4a = _mm_srai_epi32(g1x4a, in_sh);
            g1x4a = av_clip_int16_sse(g1x4a);

            // b = av_clip_int16((y * cy + cbu * u + in_rnd) >> in_sh);
            b0x4a = _mm_add_epi32(b0x4a, _mm_mullo_epi32(ux4a, _mm_set1_epi32(cbu)));
            b0x4a = _mm_add_epi32(b0x4a, rndx4);
            b0x4a = _mm_srai_epi32(b0x4a, in_sh);
            b0x4a = av_clip_int16_sse(b0x4a);

            b1x4a = _mm_add_epi32(b1x4a, _mm_mullo_epi32(ux4a, _mm_set1_epi32(cbu)));
            b1x4a = _mm_add_epi32(b1x4a, rndx4);
            b1x4a = _mm_srai_epi32(b1x4a, in_sh);
            b1x4a = av_clip_int16_sse(b1x4a);

            r0x4b = g0x4b = b0x4b = _mm_mullo_epi32(y0x4b, cyx4);
            r0x4b = _mm_add_epi32(r0x4b, _mm_mullo_epi32(vx4b, _mm_set1_epi32(crv)));
            r0x4b = _mm_add_epi32(r0x4b, rndx4);
            r0x4b = _mm_srai_epi32(r0x4b, in_sh);
            r0x4b = av_clip_int16_sse(r0x4b);

            r1x4b = g1x4b = b1x4b = _mm_mullo_epi32(y1x4b, cyx4);
            r1x4b = _mm_add_epi32(r1x4b, _mm_mullo_epi32(vx4b, _mm_set1_epi32(crv)));
            r1x4b = _mm_add_epi32(r1x4b, rndx4);
            r1x4b = _mm_srai_epi32(r1x4b, in_sh);
            r1x4b = av_clip_int16_sse(r1x4b);

            g0x4b = _mm_add_epi32(g0x4b, _mm_mullo_epi32(ux4b, _mm_set1_epi32(cgu)));
            g0x4b = _mm_add_epi32(g0x4b, _mm_mullo_epi32(vx4b, _mm_set1_epi32(cgv)));
            g0x4b = _mm_add_epi32(g0x4b, rndx4);
            g0x4b = _mm_srai_epi32(g0x4b, in_sh);
            g0x4b = av_clip_int16_sse(g0x4b);

            g1x4b = _mm_add_epi32(g1x4b, _mm_mullo_epi32(ux4b, _mm_set1_epi32(cgu)));
            g1x4b = _mm_add_epi32(g1x4b, _mm_mullo_epi32(vx4b, _mm_set1_epi32(cgv)));
            g1x4b = _mm_add_epi32(g1x4b, rndx4);
            g1x4b = _mm_srai_epi32(g1x4b, in_sh);
            g1x4b = av_clip_int16_sse(g1x4b);

            b0x4b = _mm_add_epi32(b0x4b, _mm_mullo_epi32(ux4b, _mm_set1_epi32(cbu)));
            b0x4b = _mm_add_epi32(b0x4b, rndx4);
            b0x4b = _mm_srai_epi32(b0x4b, in_sh);
            b0x4b = av_clip_int16_sse(b0x4b);

            b1x4b = _mm_add_epi32(b1x4b, _mm_mullo_epi32(ux4b, _mm_set1_epi32(cbu)));
            b1x4b = _mm_add_epi32(b1x4b, rndx4);
            b1x4b = _mm_srai_epi32(b1x4b, in_sh);
            b1x4b = av_clip_int16_sse(b1x4b);

            tonemap_int32x4_sse(r0x4a, g0x4a, b0x4a, r, g, b,
                                params->lin_lut, params->tonemap_lut, params->delin_lut,
                                params->coeffs, params->ocoeffs, params->desat, params->rgb2rgb_coeffs,
                                params->rgb2rgb_passthrough);
            tonemap_int32x4_sse(r1x4a, g1x4a, b1x4a, r1, g1, b1,
                                params->lin_lut, params->tonemap_lut, params->delin_lut,
                                params->coeffs, params->ocoeffs, params->desat, params->rgb2rgb_coeffs,
                                params->rgb2rgb_passthrough);
            tonemap_int32x4_sse(r0x4b, g0x4b, b0x4b, &r[4], &g[4], &b[4],
                                params->lin_lut, params->tonemap_lut, params->delin_lut,
                                params->coeffs, params->ocoeffs, params->desat, params->rgb2rgb_coeffs,
                                params->rgb2rgb_passthrough);
            tonemap_int32x4_sse(r1x4b, g1x4b, b1x4b, &r1[4], &g1[4], &b1[4],
                                params->lin_lut, params->tonemap_lut, params->delin_lut,
                                params->coeffs, params->ocoeffs, params->desat, params->rgb2rgb_coeffs,
                                params->rgb2rgb_passthrough);

            r0ox8 = _mm_lddqu_si128((const __m128i_u *)r);
            g0ox8 = _mm_lddqu_si128((const __m128i_u *)g);
            b0ox8 = _mm_lddqu_si128((const __m128i_u *)b);

            roax4 = _mm_cvtepi16_epi32(r0ox8);
            goax4 = _mm_cvtepi16_epi32(g0ox8);
            boax4 = _mm_cvtepi16_epi32(b0ox8);

            robx4 = _mm_unpackhi_epi16(r0ox8, zero128);
            gobx4 = _mm_unpackhi_epi16(g0ox8, zero128);
            bobx4 = _mm_unpackhi_epi16(b0ox8, zero128);

            yoax4 = _mm_mullo_epi32(roax4, _mm_set1_epi32(cry));
            yoax4 = _mm_add_epi32(yoax4, _mm_mullo_epi32(goax4, _mm_set1_epi32(cgy)));
            yoax4 = _mm_add_epi32(yoax4, _mm_mullo_epi32(boax4, _mm_set1_epi32(cby)));
            yoax4 = _mm_add_epi32(yoax4, _mm_set1_epi32(out_rnd));
            // output shift bits for 8bit outputs is 29 - 8 = 21
            yoax4 = _mm_srai_epi32(yoax4, 21);
            yoax4 = _mm_add_epi32(yoax4, _mm_set1_epi32(params->out_yuv_off));

            yobx4 = _mm_mullo_epi32(robx4, _mm_set1_epi32(cry));
            yobx4 = _mm_add_epi32(yobx4, _mm_mullo_epi32(gobx4, _mm_set1_epi32(cgy)));
            yobx4 = _mm_add_epi32(yobx4, _mm_mullo_epi32(bobx4, _mm_set1_epi32(cby)));
            yobx4 = _mm_add_epi32(yobx4, _mm_set1_epi32(out_rnd));
            yobx4 = _mm_srai_epi32(yobx4, 21);
            yobx4 = _mm_add_epi32(yobx4, _mm_set1_epi32(params->out_yuv_off));

            y0ox8 = _mm_packs_epi32(yoax4, yobx4);
            _mm_storeu_si64(&dsty[x], _mm_packus_epi16(y0ox8, zero128));

            r1ox8 = _mm_lddqu_si128((const __m128i_u *)r1);
            g1ox8 = _mm_lddqu_si128((const __m128i_u *)g1);
            b1ox8 = _mm_lddqu_si128((const __m128i_u *)b1);

            r1oax4 = _mm_cvtepi16_epi32(r1ox8);
            g1oax4 = _mm_cvtepi16_epi32(g1ox8);
            b1oax4 = _mm_cvtepi16_epi32(b1ox8);

            r1obx4 = _mm_unpackhi_epi16(r1ox8, zero128);
            g1obx4 = _mm_unpackhi_epi16(g1ox8, zero128);
            b1obx4 = _mm_unpackhi_epi16(b1ox8, zero128);

            y1oax4 = _mm_mullo_epi32(r1oax4, _mm_set1_epi32(cry));
            y1oax4 = _mm_add_epi32(y1oax4, _mm_mullo_epi32(g1oax4, _mm_set1_epi32(cgy)));
            y1oax4 = _mm_add_epi32(y1oax4, _mm_mullo_epi32(b1oax4, _mm_set1_epi32(cby)));
            y1oax4 = _mm_add_epi32(y1oax4, _mm_set1_epi32(out_rnd));
            y1oax4 = _mm_srai_epi32(y1oax4, 21);
            y1oax4 = _mm_add_epi32(y1oax4, _mm_set1_epi32(params->out_yuv_off));

            y1obx4 = _mm_mullo_epi32(r1obx4, _mm_set1_epi32(cry));
            y1obx4 = _mm_add_epi32(y1obx4, _mm_mullo_epi32(g1obx4, _mm_set1_epi32(cgy)));
            y1obx4 = _mm_add_epi32(y1obx4, _mm_mullo_epi32(b1obx4, _mm_set1_epi32(cby)));
            y1obx4 = _mm_add_epi32(y1obx4, _mm_set1_epi32(out_rnd));
            y1obx4 = _mm_srai_epi32(y1obx4, 21);
            y1obx4 = _mm_add_epi32(y1obx4, _mm_set1_epi32(params->out_yuv_off));

            y1ox8 = _mm_packs_epi32(y1oax4, y1obx4);
            _mm_storeu_si64(&dsty[x + dstlinesize[0]], _mm_packus_epi16(y1ox8, zero128));

            ravgx4 = _mm_hadd_epi32(roax4, robx4);
            ravgx4 = _mm_add_epi32(ravgx4, _mm_hadd_epi32(r1oax4, r1obx4));
            ravgx4 = _mm_add_epi32(ravgx4, _mm_set1_epi32(2));
            ravgx4 = _mm_srai_epi32(ravgx4, 2);

            gavgx4 = _mm_hadd_epi32(goax4, gobx4);
            gavgx4 = _mm_add_epi32(gavgx4, _mm_hadd_epi32(g1oax4, g1obx4));
            gavgx4 = _mm_add_epi32(gavgx4, _mm_set1_epi32(2));
            gavgx4 = _mm_srai_epi32(gavgx4, 2);

            bavgx4 = _mm_hadd_epi32(boax4, bobx4);
            bavgx4 = _mm_add_epi32(bavgx4, _mm_hadd_epi32(b1oax4, b1obx4));
            bavgx4 = _mm_add_epi32(bavgx4, _mm_set1_epi32(2));
            bavgx4 = _mm_srai_epi32(bavgx4, 2);

            uox4 = _mm_add_epi32(_mm_set1_epi32(out_rnd), _mm_mullo_epi32(ravgx4, _mm_set1_epi32(cru)));
            uox4 = _mm_add_epi32(uox4, _mm_mullo_epi32(gavgx4, _mm_set1_epi32(ocgu)));
            uox4 = _mm_add_epi32(uox4, _mm_mullo_epi32(bavgx4, _mm_set1_epi32(cburv)));
            uox4 = _mm_srai_epi32(uox4, 21);
            uox4 = _mm_add_epi32(uox4, _mm_set1_epi32(out_uv_offset));
            _mm_storeu_si32(&dstu[x >> 1], _mm_packus_epi16(_mm_packs_epi32(uox4, zero128), zero128));

            vox4 = _mm_add_epi32(_mm_set1_epi32(out_rnd), _mm_mullo_epi32(ravgx4, _mm_set1_epi32(cburv)));
            vox4 = _mm_add_epi32(vox4, _mm_mullo_epi32(gavgx4, _mm_set1_epi32(ocgv)));
            vox4 = _mm_add_epi32(vox4, _mm_mullo_epi32(bavgx4, _mm_set1_epi32(cbv)));
            vox4 = _mm_srai_epi32(vox4, 21);
            vox4 = _mm_add_epi32(vox4, _mm_set1_epi32(out_uv_offset));
            _mm_storeu_si32(&dstv[x >> 1], _mm_packus_epi16(_mm_packs_epi32(vox4, zero128), zero128));
        }
    }

    // Process remaining pixels cannot fill the full simd register with scalar version
    if (remainw) {
        int offset = width & (int)0xfffffff8;
        rdsty += offset;
        rdstu += offset >> 1;
        rdstv += offset >> 1;
        rsrcy += offset;
        rsrcu += offset >> 1;
        rsrcv += offset >> 1;
        tonemap_frame_420p10_2_420p(rdsty, rdstu, rdstv,
                                    rsrcy, rsrcu, rsrcv,
                                    dstlinesize, srclinesize,
                                    dstdepth, srcdepth,
                                    remainw, rheight, params);
    }
#endif // ENABLE_TONEMAPX_SSE_INTRINSICS
}

X86_64_V2 void tonemap_frame_420p10_2_420p10_sse(uint16_t *dsty, uint16_t *dstu, uint16_t *dstv,
                                                 const uint16_t *srcy, const uint16_t *srcu, const uint16_t *srcv,
                                                 const int *dstlinesize, const int *srclinesize,
                                                 int dstdepth, int srcdepth,
                                                 int width, int height,
                                                 const struct TonemapIntParams *params)
{
#ifdef ENABLE_TONEMAPX_SSE_INTRINSICS
    uint16_t *rdsty = dsty;
    uint16_t *rdstu = dstu;
    uint16_t *rdstv = dstv;
    const uint16_t *rsrcy = srcy;
    const uint16_t *rsrcu = srcu;
    const uint16_t *rsrcv = srcv;
    int rheight = height;
    // not zero when not divisible by 8
    // intentionally leave last pixel emtpy when input is odd
    int remainw = width & 6;

    const int in_depth = srcdepth;
    const int in_uv_offset = 128 << (in_depth - 8);
    const int in_sh = in_depth - 1;
    const int in_rnd = 1 << (in_sh - 1);

    const int out_depth = dstdepth;
    const int out_uv_offset = 128 << (out_depth - 8);
    const int out_sh = 29 - out_depth;
    const int out_rnd = 1 << (out_sh - 1);

    int cy  = (*params->yuv2rgb_coeffs)[0][0][0];
    int crv = (*params->yuv2rgb_coeffs)[0][2][0];
    int cgu = (*params->yuv2rgb_coeffs)[1][1][0];
    int cgv = (*params->yuv2rgb_coeffs)[1][2][0];
    int cbu = (*params->yuv2rgb_coeffs)[2][1][0];

    int cry   = (*params->rgb2yuv_coeffs)[0][0][0];
    int cgy   = (*params->rgb2yuv_coeffs)[0][1][0];
    int cby   = (*params->rgb2yuv_coeffs)[0][2][0];
    int cru   = (*params->rgb2yuv_coeffs)[1][0][0];
    int ocgu  = (*params->rgb2yuv_coeffs)[1][1][0];
    int cburv = (*params->rgb2yuv_coeffs)[1][2][0];
    int ocgv  = (*params->rgb2yuv_coeffs)[2][1][0];
    int cbv   = (*params->rgb2yuv_coeffs)[2][2][0];

    int16_t r[8], g[8], b[8];
    int16_t r1[8], g1[8], b1[8];

    __m128i in_yuv_offx4 = _mm_set1_epi32(params->in_yuv_off);
    __m128i in_uv_offx4= _mm_set1_epi32(in_uv_offset);
    __m128i cyx4 = _mm_set1_epi32(cy);
    __m128i rndx4 = _mm_set1_epi32(in_rnd);
    __m128i zero128 = _mm_setzero_si128();
    __m128i ux4, vx4;
    __m128i y0x8, y1x8;
    __m128i y0x4a, y0x4b, y1x4a, y1x4b, ux4a, ux4b, vx4a, vx4b;
    __m128i r0x4a, g0x4a, b0x4a, r0x4b, g0x4b, b0x4b;
    __m128i r1x4a, g1x4a, b1x4a, r1x4b, g1x4b, b1x4b;

    __m128i r0ox8, g0ox8, b0ox8;
    __m128i y0ox8;
    __m128i roax4, robx4, goax4, gobx4, boax4, bobx4;
    __m128i yoax4, yobx4;

    __m128i r1ox8, g1ox8, b1ox8;
    __m128i y1ox8;
    __m128i r1oax4, r1obx4, g1oax4, g1obx4, b1oax4, b1obx4;
    __m128i y1oax4, y1obx4;
    __m128i uox4, vox4, ravgx4, gavgx4, bavgx4;
    for (; height > 1; height -= 2,
                       dsty += dstlinesize[0], dstu += dstlinesize[1] / 2, dstv += dstlinesize[1] / 2,
                       srcy += srclinesize[0], srcu += srclinesize[1] / 2, srcv += srclinesize[1] / 2) {
        for (int xx = 0; xx < width >> 3; xx++) {
            int x = xx << 3;

            y0x8 = _mm_lddqu_si128((__m128i*)(srcy + x));
            y1x8 = _mm_lddqu_si128((__m128i*)(srcy + (srclinesize[0] / 2 + x)));
            ux4 = _mm_loadu_si64((__m128i*)(srcu + (x >> 1)));
            vx4 = _mm_loadu_si64((__m128i*)(srcv + (x >> 1)));

            y0x4a = _mm_cvtepu16_epi32(y0x8);
            y0x4b = _mm_unpackhi_epi16(y0x8, zero128);
            y1x4a = _mm_cvtepu16_epi32(y1x8);
            y1x4b = _mm_unpackhi_epi16(y1x8, zero128);
            ux4 = _mm_cvtepu16_epi32(ux4);
            vx4 = _mm_cvtepu16_epi32(vx4);
            y0x4a = _mm_sub_epi32(y0x4a, in_yuv_offx4);
            y1x4a = _mm_sub_epi32(y1x4a, in_yuv_offx4);
            y0x4b = _mm_sub_epi32(y0x4b, in_yuv_offx4);
            y1x4b = _mm_sub_epi32(y1x4b, in_yuv_offx4);
            ux4 = _mm_sub_epi32(ux4, in_uv_offx4);
            vx4 = _mm_sub_epi32(vx4, in_uv_offx4);

            ux4a = _mm_unpacklo_epi32(ux4, ux4);
            ux4b = _mm_unpackhi_epi32(ux4, ux4);
            vx4a = _mm_unpacklo_epi32(vx4, vx4);
            vx4b = _mm_unpackhi_epi32(vx4, vx4);

            // r = av_clip_int16((y * cy + crv * v + in_rnd) >> in_sh);
            r0x4a = g0x4a = b0x4a = _mm_mullo_epi32(y0x4a, cyx4);
            r0x4a = _mm_add_epi32(r0x4a, _mm_mullo_epi32(vx4a, _mm_set1_epi32(crv)));
            r0x4a = _mm_add_epi32(r0x4a, rndx4);
            r0x4a = _mm_srai_epi32(r0x4a, in_sh);
            r0x4a = av_clip_int16_sse(r0x4a);

            r1x4a = g1x4a = b1x4a = _mm_mullo_epi32(y1x4a, cyx4);
            r1x4a = _mm_add_epi32(r1x4a, _mm_mullo_epi32(vx4a, _mm_set1_epi32(crv)));
            r1x4a = _mm_add_epi32(r1x4a, rndx4);
            r1x4a = _mm_srai_epi32(r1x4a, in_sh);
            r1x4a = av_clip_int16_sse(r1x4a);

            // g = av_clip_int16((y * cy + cgu * u + cgv * v + in_rnd) >> in_sh);
            g0x4a = _mm_add_epi32(g0x4a, _mm_mullo_epi32(ux4a, _mm_set1_epi32(cgu)));
            g0x4a = _mm_add_epi32(g0x4a, _mm_mullo_epi32(vx4a, _mm_set1_epi32(cgv)));
            g0x4a = _mm_add_epi32(g0x4a, rndx4);
            g0x4a = _mm_srai_epi32(g0x4a, in_sh);
            g0x4a = av_clip_int16_sse(g0x4a);

            g1x4a = _mm_add_epi32(g1x4a, _mm_mullo_epi32(ux4a, _mm_set1_epi32(cgu)));
            g1x4a = _mm_add_epi32(g1x4a, _mm_mullo_epi32(vx4a, _mm_set1_epi32(cgv)));
            g1x4a = _mm_add_epi32(g1x4a, rndx4);
            g1x4a = _mm_srai_epi32(g1x4a, in_sh);
            g1x4a = av_clip_int16_sse(g1x4a);

            // b = av_clip_int16((y * cy + cbu * u + in_rnd) >> in_sh);
            b0x4a = _mm_add_epi32(b0x4a, _mm_mullo_epi32(ux4a, _mm_set1_epi32(cbu)));
            b0x4a = _mm_add_epi32(b0x4a, rndx4);
            b0x4a = _mm_srai_epi32(b0x4a, in_sh);
            b0x4a = av_clip_int16_sse(b0x4a);

            b1x4a = _mm_add_epi32(b1x4a, _mm_mullo_epi32(ux4a, _mm_set1_epi32(cbu)));
            b1x4a = _mm_add_epi32(b1x4a, rndx4);
            b1x4a = _mm_srai_epi32(b1x4a, in_sh);
            b1x4a = av_clip_int16_sse(b1x4a);

            r0x4b = g0x4b = b0x4b = _mm_mullo_epi32(y0x4b, cyx4);
            r0x4b = _mm_add_epi32(r0x4b, _mm_mullo_epi32(vx4b, _mm_set1_epi32(crv)));
            r0x4b = _mm_add_epi32(r0x4b, rndx4);
            r0x4b = _mm_srai_epi32(r0x4b, in_sh);
            r0x4b = av_clip_int16_sse(r0x4b);

            r1x4b = g1x4b = b1x4b = _mm_mullo_epi32(y1x4b, cyx4);
            r1x4b = _mm_add_epi32(r1x4b, _mm_mullo_epi32(vx4b, _mm_set1_epi32(crv)));
            r1x4b = _mm_add_epi32(r1x4b, rndx4);
            r1x4b = _mm_srai_epi32(r1x4b, in_sh);
            r1x4b = av_clip_int16_sse(r1x4b);

            g0x4b = _mm_add_epi32(g0x4b, _mm_mullo_epi32(ux4b, _mm_set1_epi32(cgu)));
            g0x4b = _mm_add_epi32(g0x4b, _mm_mullo_epi32(vx4b, _mm_set1_epi32(cgv)));
            g0x4b = _mm_add_epi32(g0x4b, rndx4);
            g0x4b = _mm_srai_epi32(g0x4b, in_sh);
            g0x4b = av_clip_int16_sse(g0x4b);

            g1x4b = _mm_add_epi32(g1x4b, _mm_mullo_epi32(ux4b, _mm_set1_epi32(cgu)));
            g1x4b = _mm_add_epi32(g1x4b, _mm_mullo_epi32(vx4b, _mm_set1_epi32(cgv)));
            g1x4b = _mm_add_epi32(g1x4b, rndx4);
            g1x4b = _mm_srai_epi32(g1x4b, in_sh);
            g1x4b = av_clip_int16_sse(g1x4b);

            b0x4b = _mm_add_epi32(b0x4b, _mm_mullo_epi32(ux4b, _mm_set1_epi32(cbu)));
            b0x4b = _mm_add_epi32(b0x4b, rndx4);
            b0x4b = _mm_srai_epi32(b0x4b, in_sh);
            b0x4b = av_clip_int16_sse(b0x4b);

            b1x4b = _mm_add_epi32(b1x4b, _mm_mullo_epi32(ux4b, _mm_set1_epi32(cbu)));
            b1x4b = _mm_add_epi32(b1x4b, rndx4);
            b1x4b = _mm_srai_epi32(b1x4b, in_sh);
            b1x4b = av_clip_int16_sse(b1x4b);

            tonemap_int32x4_sse(r0x4a, g0x4a, b0x4a, r, g, b,
                                params->lin_lut, params->tonemap_lut, params->delin_lut,
                                params->coeffs, params->ocoeffs, params->desat, params->rgb2rgb_coeffs,
                                params->rgb2rgb_passthrough);
            tonemap_int32x4_sse(r1x4a, g1x4a, b1x4a, r1, g1, b1,
                                params->lin_lut, params->tonemap_lut, params->delin_lut,
                                params->coeffs, params->ocoeffs, params->desat, params->rgb2rgb_coeffs,
                                params->rgb2rgb_passthrough);
            tonemap_int32x4_sse(r0x4b, g0x4b, b0x4b, &r[4], &g[4], &b[4],
                                params->lin_lut, params->tonemap_lut, params->delin_lut,
                                params->coeffs, params->ocoeffs, params->desat, params->rgb2rgb_coeffs,
                                params->rgb2rgb_passthrough);
            tonemap_int32x4_sse(r1x4b, g1x4b, b1x4b, &r1[4], &g1[4], &b1[4],
                                params->lin_lut, params->tonemap_lut, params->delin_lut,
                                params->coeffs, params->ocoeffs, params->desat, params->rgb2rgb_coeffs,
                                params->rgb2rgb_passthrough);

            r0ox8 = _mm_lddqu_si128((const __m128i_u *)r);
            g0ox8 = _mm_lddqu_si128((const __m128i_u *)g);
            b0ox8 = _mm_lddqu_si128((const __m128i_u *)b);

            roax4 = _mm_cvtepi16_epi32(r0ox8);
            goax4 = _mm_cvtepi16_epi32(g0ox8);
            boax4 = _mm_cvtepi16_epi32(b0ox8);

            robx4 = _mm_unpackhi_epi16(r0ox8, zero128);
            gobx4 = _mm_unpackhi_epi16(g0ox8, zero128);
            bobx4 = _mm_unpackhi_epi16(b0ox8, zero128);

            yoax4 = _mm_mullo_epi32(roax4, _mm_set1_epi32(cry));
            yoax4 = _mm_add_epi32(yoax4, _mm_mullo_epi32(goax4, _mm_set1_epi32(cgy)));
            yoax4 = _mm_add_epi32(yoax4, _mm_mullo_epi32(boax4, _mm_set1_epi32(cby)));
            yoax4 = _mm_add_epi32(yoax4, _mm_set1_epi32(out_rnd));
            yoax4 = _mm_srai_epi32(yoax4, out_sh);
            yoax4 = _mm_add_epi32(yoax4, _mm_set1_epi32(params->out_yuv_off));

            yobx4 = _mm_mullo_epi32(robx4, _mm_set1_epi32(cry));
            yobx4 = _mm_add_epi32(yobx4, _mm_mullo_epi32(gobx4, _mm_set1_epi32(cgy)));
            yobx4 = _mm_add_epi32(yobx4, _mm_mullo_epi32(bobx4, _mm_set1_epi32(cby)));
            yobx4 = _mm_add_epi32(yobx4, _mm_set1_epi32(out_rnd));
            yobx4 = _mm_srai_epi32(yobx4, out_sh);
            yobx4 = _mm_add_epi32(yobx4, _mm_set1_epi32(params->out_yuv_off));

            y0ox8 = _mm_packus_epi32(yoax4, yobx4);
            _mm_storeu_si128((__m128i_u *) &dsty[x], y0ox8);

            r1ox8 = _mm_lddqu_si128((const __m128i_u *)r1);
            g1ox8 = _mm_lddqu_si128((const __m128i_u *)g1);
            b1ox8 = _mm_lddqu_si128((const __m128i_u *)b1);

            r1oax4 = _mm_cvtepi16_epi32(r1ox8);
            g1oax4 = _mm_cvtepi16_epi32(g1ox8);
            b1oax4 = _mm_cvtepi16_epi32(b1ox8);

            r1obx4 = _mm_unpackhi_epi16(r1ox8, zero128);
            g1obx4 = _mm_unpackhi_epi16(g1ox8, zero128);
            b1obx4 = _mm_unpackhi_epi16(b1ox8, zero128);

            y1oax4 = _mm_mullo_epi32(r1oax4, _mm_set1_epi32(cry));
            y1oax4 = _mm_add_epi32(y1oax4, _mm_mullo_epi32(g1oax4, _mm_set1_epi32(cgy)));
            y1oax4 = _mm_add_epi32(y1oax4, _mm_mullo_epi32(b1oax4, _mm_set1_epi32(cby)));
            y1oax4 = _mm_add_epi32(y1oax4, _mm_set1_epi32(out_rnd));
            y1oax4 = _mm_srai_epi32(y1oax4, out_sh);
            y1oax4 = _mm_add_epi32(y1oax4, _mm_set1_epi32(params->out_yuv_off));

            y1obx4 = _mm_mullo_epi32(r1obx4, _mm_set1_epi32(cry));
            y1obx4 = _mm_add_epi32(y1obx4, _mm_mullo_epi32(g1obx4, _mm_set1_epi32(cgy)));
            y1obx4 = _mm_add_epi32(y1obx4, _mm_mullo_epi32(b1obx4, _mm_set1_epi32(cby)));
            y1obx4 = _mm_add_epi32(y1obx4, _mm_set1_epi32(out_rnd));
            y1obx4 = _mm_srai_epi32(y1obx4, out_sh);
            y1obx4 = _mm_add_epi32(y1obx4, _mm_set1_epi32(params->out_yuv_off));

            y1ox8 = _mm_packus_epi32(y1oax4, y1obx4);
            _mm_storeu_si128((__m128i_u *) &dsty[x + dstlinesize[0] / 2], y1ox8);

            ravgx4 = _mm_hadd_epi32(roax4, robx4);
            ravgx4 = _mm_add_epi32(ravgx4, _mm_hadd_epi32(r1oax4, r1obx4));
            ravgx4 = _mm_add_epi32(ravgx4, _mm_set1_epi32(2));
            ravgx4 = _mm_srai_epi32(ravgx4, 2);

            gavgx4 = _mm_hadd_epi32(goax4, gobx4);
            gavgx4 = _mm_add_epi32(gavgx4, _mm_hadd_epi32(g1oax4, g1obx4));
            gavgx4 = _mm_add_epi32(gavgx4, _mm_set1_epi32(2));
            gavgx4 = _mm_srai_epi32(gavgx4, 2);

            bavgx4 = _mm_hadd_epi32(boax4, bobx4);
            bavgx4 = _mm_add_epi32(bavgx4, _mm_hadd_epi32(b1oax4, b1obx4));
            bavgx4 = _mm_add_epi32(bavgx4, _mm_set1_epi32(2));
            bavgx4 = _mm_srai_epi32(bavgx4, 2);

            uox4 = _mm_add_epi32(_mm_set1_epi32(out_rnd), _mm_mullo_epi32(ravgx4, _mm_set1_epi32(cru)));
            uox4 = _mm_add_epi32(uox4, _mm_mullo_epi32(gavgx4, _mm_set1_epi32(ocgu)));
            uox4 = _mm_add_epi32(uox4, _mm_mullo_epi32(bavgx4, _mm_set1_epi32(cburv)));
            uox4 = _mm_srai_epi32(uox4, out_sh);
            uox4 = _mm_add_epi32(uox4, _mm_set1_epi32(out_uv_offset));
            _mm_storeu_si64((__m128i_u *) &dstu[x >> 1], _mm_packus_epi32(uox4, zero128));

            vox4 = _mm_add_epi32(_mm_set1_epi32(out_rnd), _mm_mullo_epi32(ravgx4, _mm_set1_epi32(cburv)));
            vox4 = _mm_add_epi32(vox4, _mm_mullo_epi32(gavgx4, _mm_set1_epi32(ocgv)));
            vox4 = _mm_add_epi32(vox4, _mm_mullo_epi32(bavgx4, _mm_set1_epi32(cbv)));
            vox4 = _mm_srai_epi32(vox4, out_sh);
            vox4 = _mm_add_epi32(vox4, _mm_set1_epi32(out_uv_offset));
            _mm_storeu_si64((__m128i_u *) &dstv[x >> 1], _mm_packus_epi32(vox4, zero128));
        }
    }

    // Process remaining pixels cannot fill the full simd register with scalar version
    if (remainw) {
        int offset = width & (int)0xfffffff8;
        rdsty += offset;
        rdstu += offset >> 1;
        rdstv += offset >> 1;
        rsrcy += offset;
        rsrcu += offset >> 1;
        rsrcv += offset >> 1;
        tonemap_frame_420p10_2_420p10(rdsty, rdstu, rdstv,
                                      rsrcy, rsrcu, rsrcv,
                                      dstlinesize, srclinesize,
                                      dstdepth, srcdepth,
                                      remainw, rheight, params);
    }
#endif // ENABLE_TONEMAPX_SSE_INTRINSICS
}

X86_64_V2 void tonemap_frame_p016_p010_2_nv12_sse(uint8_t *dsty, uint8_t *dstuv,
                                                  const uint16_t *srcy, const uint16_t *srcuv,
                                                  const int *dstlinesize, const int *srclinesize,
                                                  int dstdepth, int srcdepth,
                                                  int width, int height,
                                                  const struct TonemapIntParams *params)
{
#ifdef ENABLE_TONEMAPX_SSE_INTRINSICS
    uint8_t *rdsty = dsty;
    uint8_t *rdstuv = dstuv;
    const uint16_t *rsrcy = srcy;
    const uint16_t *rsrcuv = srcuv;
    int rheight = height;
    // not zero when not divisible by 8
    // intentionally leave last pixel emtpy when input is odd
    int remainw = width & 6;

    const int in_depth = srcdepth;
    const int in_uv_offset = 128 << (in_depth - 8);
    const int in_sh = in_depth - 1;
    const int in_rnd = 1 << (in_sh - 1);

    const int out_depth = dstdepth;
    const int out_uv_offset = 128 << (out_depth - 8);
    const int out_sh = 29 - out_depth;
    const int out_rnd = 1 << (out_sh - 1);

    int cy  = (*params->yuv2rgb_coeffs)[0][0][0];
    int crv = (*params->yuv2rgb_coeffs)[0][2][0];
    int cgu = (*params->yuv2rgb_coeffs)[1][1][0];
    int cgv = (*params->yuv2rgb_coeffs)[1][2][0];
    int cbu = (*params->yuv2rgb_coeffs)[2][1][0];

    int cry   = (*params->rgb2yuv_coeffs)[0][0][0];
    int cgy   = (*params->rgb2yuv_coeffs)[0][1][0];
    int cby   = (*params->rgb2yuv_coeffs)[0][2][0];
    int cru   = (*params->rgb2yuv_coeffs)[1][0][0];
    int ocgu  = (*params->rgb2yuv_coeffs)[1][1][0];
    int cburv = (*params->rgb2yuv_coeffs)[1][2][0];
    int ocgv  = (*params->rgb2yuv_coeffs)[2][1][0];
    int cbv   = (*params->rgb2yuv_coeffs)[2][2][0];

    int16_t r[8], g[8], b[8];
    int16_t r1[8], g1[8], b1[8];

    __m128i in_yuv_offx4 = _mm_set1_epi32(params->in_yuv_off);
    __m128i in_uv_offx4= _mm_set1_epi32(in_uv_offset);
    __m128i cyx4 = _mm_set1_epi32(cy);
    __m128i rndx4 = _mm_set1_epi32(in_rnd);
    __m128i zero128 = _mm_setzero_si128();
    __m128i uvx8, uvx4a, uvx4b;
    __m128i y0x8, y1x8;
    __m128i y0x4a, y0x4b, y1x4a, y1x4b, ux4a, ux4b, vx4a, vx4b;
    __m128i r0x4a, g0x4a, b0x4a, r0x4b, g0x4b, b0x4b;
    __m128i r1x4a, g1x4a, b1x4a, r1x4b, g1x4b, b1x4b;

    __m128i r0ox8, g0ox8, b0ox8;
    __m128i y0ox8;
    __m128i roax4, robx4, goax4, gobx4, boax4, bobx4;
    __m128i yoax4, yobx4;

    __m128i r1ox8, g1ox8, b1ox8;
    __m128i y1ox8;
    __m128i r1oax4, r1obx4, g1oax4, g1obx4, b1oax4, b1obx4;
    __m128i y1oax4, y1obx4, uvoax4, uvobx4;
    __m128i uoax4, voax4, ravgx4, gavgx4, bavgx4;
    for (; height > 1; height -= 2,
                       dsty += dstlinesize[0] * 2, dstuv += dstlinesize[1],
                       srcy += srclinesize[0], srcuv += srclinesize[1] / 2) {
        for (int xx = 0; xx < width >> 3; xx++) {
            int x = xx << 3;

            y0x8 = _mm_lddqu_si128((__m128i*)(srcy + x));
            y1x8 = _mm_lddqu_si128((__m128i*)(srcy + (srclinesize[0] / 2 + x)));
            uvx8 = _mm_lddqu_si128((__m128i*)(srcuv + x));

            if (in_depth == 10) {
                // shift to low10bits for 10bit input
                // shift bit has to be compile-time constant
                y0x8 = _mm_srli_epi16(y0x8, 6);
                y1x8 = _mm_srli_epi16(y1x8, 6);
                uvx8 = _mm_srli_epi16(uvx8, 6);
            }
            y0x4a = _mm_cvtepu16_epi32(y0x8);
            y0x4b = _mm_unpackhi_epi16(y0x8, zero128);
            y1x4a = _mm_cvtepu16_epi32(y1x8);
            y1x4b = _mm_unpackhi_epi16(y1x8, zero128);
            uvx4a = _mm_cvtepu16_epi32(uvx8);
            uvx4b = _mm_unpackhi_epi16(uvx8, zero128);
            y0x4a = _mm_sub_epi32(y0x4a, in_yuv_offx4);
            y1x4a = _mm_sub_epi32(y1x4a, in_yuv_offx4);
            y0x4b = _mm_sub_epi32(y0x4b, in_yuv_offx4);
            y1x4b = _mm_sub_epi32(y1x4b, in_yuv_offx4);
            uvx4a = _mm_sub_epi32(uvx4a, in_uv_offx4);
            uvx4b = _mm_sub_epi32(uvx4b, in_uv_offx4);

            ux4a = _mm_shuffle_epi32(uvx4a, _MM_SHUFFLE(2, 2, 0, 0));
            ux4b = _mm_shuffle_epi32(uvx4b, _MM_SHUFFLE(2, 2, 0, 0));
            vx4a = _mm_shuffle_epi32(uvx4a, _MM_SHUFFLE(3, 3, 1, 1));
            vx4b = _mm_shuffle_epi32(uvx4b, _MM_SHUFFLE(3, 3, 1, 1));

            // r = av_clip_int16((y * cy + crv * v + in_rnd) >> in_sh);
            r0x4a = g0x4a = b0x4a = _mm_mullo_epi32(y0x4a, cyx4);
            r0x4a = _mm_add_epi32(r0x4a, _mm_mullo_epi32(vx4a, _mm_set1_epi32(crv)));
            r0x4a = _mm_add_epi32(r0x4a, rndx4);
            r0x4a = _mm_srai_epi32(r0x4a, in_sh);
            r0x4a = av_clip_int16_sse(r0x4a);

            r1x4a = g1x4a = b1x4a = _mm_mullo_epi32(y1x4a, cyx4);
            r1x4a = _mm_add_epi32(r1x4a, _mm_mullo_epi32(vx4a, _mm_set1_epi32(crv)));
            r1x4a = _mm_add_epi32(r1x4a, rndx4);
            r1x4a = _mm_srai_epi32(r1x4a, in_sh);
            r1x4a = av_clip_int16_sse(r1x4a);

            // g = av_clip_int16((y * cy + cgu * u + cgv * v + in_rnd) >> in_sh);
            g0x4a = _mm_add_epi32(g0x4a, _mm_mullo_epi32(ux4a, _mm_set1_epi32(cgu)));
            g0x4a = _mm_add_epi32(g0x4a, _mm_mullo_epi32(vx4a, _mm_set1_epi32(cgv)));
            g0x4a = _mm_add_epi32(g0x4a, rndx4);
            g0x4a = _mm_srai_epi32(g0x4a, in_sh);
            g0x4a = av_clip_int16_sse(g0x4a);

            g1x4a = _mm_add_epi32(g1x4a, _mm_mullo_epi32(ux4a, _mm_set1_epi32(cgu)));
            g1x4a = _mm_add_epi32(g1x4a, _mm_mullo_epi32(vx4a, _mm_set1_epi32(cgv)));
            g1x4a = _mm_add_epi32(g1x4a, rndx4);
            g1x4a = _mm_srai_epi32(g1x4a, in_sh);
            g1x4a = av_clip_int16_sse(g1x4a);

            // b = av_clip_int16((y * cy + cbu * u + in_rnd) >> in_sh);
            b0x4a = _mm_add_epi32(b0x4a, _mm_mullo_epi32(ux4a, _mm_set1_epi32(cbu)));
            b0x4a = _mm_add_epi32(b0x4a, rndx4);
            b0x4a = _mm_srai_epi32(b0x4a, in_sh);
            b0x4a = av_clip_int16_sse(b0x4a);

            b1x4a = _mm_add_epi32(b1x4a, _mm_mullo_epi32(ux4a, _mm_set1_epi32(cbu)));
            b1x4a = _mm_add_epi32(b1x4a, rndx4);
            b1x4a = _mm_srai_epi32(b1x4a, in_sh);
            b1x4a = av_clip_int16_sse(b1x4a);

            r0x4b = g0x4b = b0x4b = _mm_mullo_epi32(y0x4b, cyx4);
            r0x4b = _mm_add_epi32(r0x4b, _mm_mullo_epi32(vx4b, _mm_set1_epi32(crv)));
            r0x4b = _mm_add_epi32(r0x4b, rndx4);
            r0x4b = _mm_srai_epi32(r0x4b, in_sh);
            r0x4b = av_clip_int16_sse(r0x4b);

            r1x4b = g1x4b = b1x4b = _mm_mullo_epi32(y1x4b, cyx4);
            r1x4b = _mm_add_epi32(r1x4b, _mm_mullo_epi32(vx4b, _mm_set1_epi32(crv)));
            r1x4b = _mm_add_epi32(r1x4b, rndx4);
            r1x4b = _mm_srai_epi32(r1x4b, in_sh);
            r1x4b = av_clip_int16_sse(r1x4b);

            g0x4b = _mm_add_epi32(g0x4b, _mm_mullo_epi32(ux4b, _mm_set1_epi32(cgu)));
            g0x4b = _mm_add_epi32(g0x4b, _mm_mullo_epi32(vx4b, _mm_set1_epi32(cgv)));
            g0x4b = _mm_add_epi32(g0x4b, rndx4);
            g0x4b = _mm_srai_epi32(g0x4b, in_sh);
            g0x4b = av_clip_int16_sse(g0x4b);

            g1x4b = _mm_add_epi32(g1x4b, _mm_mullo_epi32(ux4b, _mm_set1_epi32(cgu)));
            g1x4b = _mm_add_epi32(g1x4b, _mm_mullo_epi32(vx4b, _mm_set1_epi32(cgv)));
            g1x4b = _mm_add_epi32(g1x4b, rndx4);
            g1x4b = _mm_srai_epi32(g1x4b, in_sh);
            g1x4b = av_clip_int16_sse(g1x4b);

            b0x4b = _mm_add_epi32(b0x4b, _mm_mullo_epi32(ux4b, _mm_set1_epi32(cbu)));
            b0x4b = _mm_add_epi32(b0x4b, rndx4);
            b0x4b = _mm_srai_epi32(b0x4b, in_sh);
            b0x4b = av_clip_int16_sse(b0x4b);

            b1x4b = _mm_add_epi32(b1x4b, _mm_mullo_epi32(ux4b, _mm_set1_epi32(cbu)));
            b1x4b = _mm_add_epi32(b1x4b, rndx4);
            b1x4b = _mm_srai_epi32(b1x4b, in_sh);
            b1x4b = av_clip_int16_sse(b1x4b);

            tonemap_int32x4_sse(r0x4a, g0x4a, b0x4a, r, g, b,
                                params->lin_lut, params->tonemap_lut, params->delin_lut,
                                params->coeffs, params->ocoeffs, params->desat, params->rgb2rgb_coeffs,
                                params->rgb2rgb_passthrough);
            tonemap_int32x4_sse(r1x4a, g1x4a, b1x4a, r1, g1, b1,
                                params->lin_lut, params->tonemap_lut, params->delin_lut,
                                params->coeffs, params->ocoeffs, params->desat, params->rgb2rgb_coeffs,
                                params->rgb2rgb_passthrough);
            tonemap_int32x4_sse(r0x4b, g0x4b, b0x4b, &r[4], &g[4], &b[4],
                                params->lin_lut, params->tonemap_lut, params->delin_lut,
                                params->coeffs, params->ocoeffs, params->desat, params->rgb2rgb_coeffs,
                                params->rgb2rgb_passthrough);
            tonemap_int32x4_sse(r1x4b, g1x4b, b1x4b, &r1[4], &g1[4], &b1[4],
                                params->lin_lut, params->tonemap_lut, params->delin_lut,
                                params->coeffs, params->ocoeffs, params->desat, params->rgb2rgb_coeffs,
                                params->rgb2rgb_passthrough);

            r0ox8 = _mm_lddqu_si128((const __m128i_u *)r);
            g0ox8 = _mm_lddqu_si128((const __m128i_u *)g);
            b0ox8 = _mm_lddqu_si128((const __m128i_u *)b);

            roax4 = _mm_cvtepi16_epi32(r0ox8);
            goax4 = _mm_cvtepi16_epi32(g0ox8);
            boax4 = _mm_cvtepi16_epi32(b0ox8);

            robx4 = _mm_unpackhi_epi16(r0ox8, zero128);
            gobx4 = _mm_unpackhi_epi16(g0ox8, zero128);
            bobx4 = _mm_unpackhi_epi16(b0ox8, zero128);

            yoax4 = _mm_mullo_epi32(roax4, _mm_set1_epi32(cry));
            yoax4 = _mm_add_epi32(yoax4, _mm_mullo_epi32(goax4, _mm_set1_epi32(cgy)));
            yoax4 = _mm_add_epi32(yoax4, _mm_mullo_epi32(boax4, _mm_set1_epi32(cby)));
            yoax4 = _mm_add_epi32(yoax4, _mm_set1_epi32(out_rnd));
            // output shift bits for 8bit outputs is 29 - 8 = 21
            yoax4 = _mm_srai_epi32(yoax4, 21);
            yoax4 = _mm_add_epi32(yoax4, _mm_set1_epi32(params->out_yuv_off));

            yobx4 = _mm_mullo_epi32(robx4, _mm_set1_epi32(cry));
            yobx4 = _mm_add_epi32(yobx4, _mm_mullo_epi32(gobx4, _mm_set1_epi32(cgy)));
            yobx4 = _mm_add_epi32(yobx4, _mm_mullo_epi32(bobx4, _mm_set1_epi32(cby)));
            yobx4 = _mm_add_epi32(yobx4, _mm_set1_epi32(out_rnd));
            yobx4 = _mm_srai_epi32(yobx4, 21);
            yobx4 = _mm_add_epi32(yobx4, _mm_set1_epi32(params->out_yuv_off));

            y0ox8 = _mm_packs_epi32(yoax4, yobx4);
            _mm_storeu_si64(&dsty[x], _mm_packus_epi16(y0ox8, zero128));

            r1ox8 = _mm_lddqu_si128((const __m128i_u *)r1);
            g1ox8 = _mm_lddqu_si128((const __m128i_u *)g1);
            b1ox8 = _mm_lddqu_si128((const __m128i_u *)b1);

            r1oax4 = _mm_cvtepi16_epi32(r1ox8);
            g1oax4 = _mm_cvtepi16_epi32(g1ox8);
            b1oax4 = _mm_cvtepi16_epi32(b1ox8);

            r1obx4 = _mm_unpackhi_epi16(r1ox8, zero128);
            g1obx4 = _mm_unpackhi_epi16(g1ox8, zero128);
            b1obx4 = _mm_unpackhi_epi16(b1ox8, zero128);

            y1oax4 = _mm_mullo_epi32(r1oax4, _mm_set1_epi32(cry));
            y1oax4 = _mm_add_epi32(y1oax4, _mm_mullo_epi32(g1oax4, _mm_set1_epi32(cgy)));
            y1oax4 = _mm_add_epi32(y1oax4, _mm_mullo_epi32(b1oax4, _mm_set1_epi32(cby)));
            y1oax4 = _mm_add_epi32(y1oax4, _mm_set1_epi32(out_rnd));
            y1oax4 = _mm_srai_epi32(y1oax4, 21);
            y1oax4 = _mm_add_epi32(y1oax4, _mm_set1_epi32(params->out_yuv_off));

            y1obx4 = _mm_mullo_epi32(r1obx4, _mm_set1_epi32(cry));
            y1obx4 = _mm_add_epi32(y1obx4, _mm_mullo_epi32(g1obx4, _mm_set1_epi32(cgy)));
            y1obx4 = _mm_add_epi32(y1obx4, _mm_mullo_epi32(b1obx4, _mm_set1_epi32(cby)));
            y1obx4 = _mm_add_epi32(y1obx4, _mm_set1_epi32(out_rnd));
            y1obx4 = _mm_srai_epi32(y1obx4, 21);
            y1obx4 = _mm_add_epi32(y1obx4, _mm_set1_epi32(params->out_yuv_off));

            y1ox8 = _mm_packs_epi32(y1oax4, y1obx4);
            _mm_storeu_si64(&dsty[x + dstlinesize[0]], _mm_packus_epi16(y1ox8, zero128));

            ravgx4 = _mm_hadd_epi32(roax4, robx4);
            ravgx4 = _mm_add_epi32(ravgx4, _mm_hadd_epi32(r1oax4, r1obx4));
            ravgx4 = _mm_add_epi32(ravgx4, _mm_set1_epi32(2));
            ravgx4 = _mm_srai_epi32(ravgx4, 2);

            gavgx4 = _mm_hadd_epi32(goax4, gobx4);
            gavgx4 = _mm_add_epi32(gavgx4, _mm_hadd_epi32(g1oax4, g1obx4));
            gavgx4 = _mm_add_epi32(gavgx4, _mm_set1_epi32(2));
            gavgx4 = _mm_srai_epi32(gavgx4, 2);

            bavgx4 = _mm_hadd_epi32(boax4, bobx4);
            bavgx4 = _mm_add_epi32(bavgx4, _mm_hadd_epi32(b1oax4, b1obx4));
            bavgx4 = _mm_add_epi32(bavgx4, _mm_set1_epi32(2));
            bavgx4 = _mm_srai_epi32(bavgx4, 2);

            uoax4 = _mm_add_epi32(_mm_set1_epi32(out_rnd), _mm_mullo_epi32(ravgx4, _mm_set1_epi32(cru)));
            uoax4 = _mm_add_epi32(uoax4, _mm_mullo_epi32(gavgx4, _mm_set1_epi32(ocgu)));
            uoax4 = _mm_add_epi32(uoax4, _mm_mullo_epi32(bavgx4, _mm_set1_epi32(cburv)));
            uoax4 = _mm_srai_epi32(uoax4, 21);
            uoax4 = _mm_add_epi32(uoax4, _mm_set1_epi32(out_uv_offset));

            voax4 = _mm_add_epi32(_mm_set1_epi32(out_rnd), _mm_mullo_epi32(ravgx4, _mm_set1_epi32(cburv)));
            voax4 = _mm_add_epi32(voax4, _mm_mullo_epi32(gavgx4, _mm_set1_epi32(ocgv)));
            voax4 = _mm_add_epi32(voax4, _mm_mullo_epi32(bavgx4, _mm_set1_epi32(cbv)));
            voax4 = _mm_srai_epi32(voax4, 21);
            voax4 = _mm_add_epi32(voax4, _mm_set1_epi32(out_uv_offset));

            uvoax4 = _mm_unpacklo_epi32(uoax4, voax4);
            uvobx4 = _mm_unpackhi_epi32(uoax4, voax4);
            _mm_storeu_si64(&dstuv[x], _mm_packus_epi16(_mm_packs_epi32(uvoax4, uvobx4), zero128));
        }
    }

    // Process remaining pixels cannot fill the full simd register with scalar version
    if (remainw) {
        int offset = width & (int)0xfffffff8;
        rdsty += offset;
        rdstuv += offset;
        rsrcy += offset;
        rsrcuv += offset;
        tonemap_frame_p016_p010_2_nv12(rdsty, rdstuv,
                                       rsrcy, rsrcuv,
                                       dstlinesize, srclinesize,
                                       dstdepth, srcdepth,
                                       remainw, rheight, params);
    }
#endif // ENABLE_TONEMAPX_SSE_INTRINSICS
}

X86_64_V2 void tonemap_frame_p016_p010_2_p016_p010_sse(uint16_t *dsty, uint16_t *dstuv,
                                                       const uint16_t *srcy, const uint16_t *srcuv,
                                                       const int *dstlinesize, const int *srclinesize,
                                                       int dstdepth, int srcdepth,
                                                       int width, int height,
                                                       const struct TonemapIntParams *params)
{
#ifdef ENABLE_TONEMAPX_SSE_INTRINSICS
    uint16_t *rdsty = dsty;
    uint16_t *rdstuv = dstuv;
    const uint16_t *rsrcy = srcy;
    const uint16_t *rsrcuv = srcuv;
    int rheight = height;
    // not zero when not divisible by 8
    // intentionally leave last pixel emtpy when input is odd
    int remainw = width & 6;

    const int in_depth = srcdepth;
    const int in_uv_offset = 128 << (in_depth - 8);
    const int in_sh = in_depth - 1;
    const int in_rnd = 1 << (in_sh - 1);

    const int out_depth = dstdepth;
    const int out_uv_offset = 128 << (out_depth - 8);
    const int out_sh = 29 - out_depth;
    const int out_rnd = 1 << (out_sh - 1);
    const int out_sh2 = 16 - out_depth;

    int cy  = (*params->yuv2rgb_coeffs)[0][0][0];
    int crv = (*params->yuv2rgb_coeffs)[0][2][0];
    int cgu = (*params->yuv2rgb_coeffs)[1][1][0];
    int cgv = (*params->yuv2rgb_coeffs)[1][2][0];
    int cbu = (*params->yuv2rgb_coeffs)[2][1][0];

    int cry   = (*params->rgb2yuv_coeffs)[0][0][0];
    int cgy   = (*params->rgb2yuv_coeffs)[0][1][0];
    int cby   = (*params->rgb2yuv_coeffs)[0][2][0];
    int cru   = (*params->rgb2yuv_coeffs)[1][0][0];
    int ocgu  = (*params->rgb2yuv_coeffs)[1][1][0];
    int cburv = (*params->rgb2yuv_coeffs)[1][2][0];
    int ocgv  = (*params->rgb2yuv_coeffs)[2][1][0];
    int cbv   = (*params->rgb2yuv_coeffs)[2][2][0];

    int16_t r[8], g[8], b[8];
    int16_t r1[8], g1[8], b1[8];

    __m128i in_yuv_offx4 = _mm_set1_epi32(params->in_yuv_off);
    __m128i in_uv_offx4= _mm_set1_epi32(in_uv_offset);
    __m128i cyx4 = _mm_set1_epi32(cy);
    __m128i rndx4 = _mm_set1_epi32(in_rnd);
    __m128i zero128 = _mm_setzero_si128();
    __m128i uvx8, uvx4a, uvx4b;
    __m128i y0x8, y1x8;
    __m128i y0x4a, y0x4b, y1x4a, y1x4b, ux4a, ux4b, vx4a, vx4b;
    __m128i r0x4a, g0x4a, b0x4a, r0x4b, g0x4b, b0x4b;
    __m128i r1x4a, g1x4a, b1x4a, r1x4b, g1x4b, b1x4b;

    __m128i r0ox8, g0ox8, b0ox8;
    __m128i y0ox8;
    __m128i roax4, robx4, goax4, gobx4, boax4, bobx4;
    __m128i yoax4, yobx4;

    __m128i r1ox8, g1ox8, b1ox8;
    __m128i y1ox8;
    __m128i r1oax4, r1obx4, g1oax4, g1obx4, b1oax4, b1obx4;
    __m128i y1oax4, y1obx4, uvoax4, uvobx4;
    __m128i uoax4, voax4, ravgx4, gavgx4, bavgx4, uvox8;
    for (; height > 1; height -= 2,
                       dsty += dstlinesize[0], dstuv += dstlinesize[1] / 2,
                       srcy += srclinesize[0], srcuv += srclinesize[1] / 2) {
        for (int xx = 0; xx < width >> 3; xx++) {
            int x = xx << 3;

            y0x8 = _mm_lddqu_si128((__m128i*)(srcy + x));
            y1x8 = _mm_lddqu_si128((__m128i*)(srcy + (srclinesize[0] / 2 + x)));
            uvx8 = _mm_lddqu_si128((__m128i*)(srcuv + x));

            if (in_depth == 10) {
                // shift to low10bits for 10bit input
                // shift bit has to be compile-time constant
                y0x8 = _mm_srli_epi16(y0x8, 6);
                y1x8 = _mm_srli_epi16(y1x8, 6);
                uvx8 = _mm_srli_epi16(uvx8, 6);
            }
            y0x4a = _mm_cvtepu16_epi32(y0x8);
            y0x4b = _mm_unpackhi_epi16(y0x8, zero128);
            y1x4a = _mm_cvtepu16_epi32(y1x8);
            y1x4b = _mm_unpackhi_epi16(y1x8, zero128);
            uvx4a = _mm_cvtepu16_epi32(uvx8);
            uvx4b = _mm_unpackhi_epi16(uvx8, zero128);
            y0x4a = _mm_sub_epi32(y0x4a, in_yuv_offx4);
            y1x4a = _mm_sub_epi32(y1x4a, in_yuv_offx4);
            y0x4b = _mm_sub_epi32(y0x4b, in_yuv_offx4);
            y1x4b = _mm_sub_epi32(y1x4b, in_yuv_offx4);
            uvx4a = _mm_sub_epi32(uvx4a, in_uv_offx4);
            uvx4b = _mm_sub_epi32(uvx4b, in_uv_offx4);

            ux4a = _mm_shuffle_epi32(uvx4a, _MM_SHUFFLE(2, 2, 0, 0));
            ux4b = _mm_shuffle_epi32(uvx4b, _MM_SHUFFLE(2, 2, 0, 0));
            vx4a = _mm_shuffle_epi32(uvx4a, _MM_SHUFFLE(3, 3, 1, 1));
            vx4b = _mm_shuffle_epi32(uvx4b, _MM_SHUFFLE(3, 3, 1, 1));

            // r = av_clip_int16((y * cy + crv * v + in_rnd) >> in_sh);
            r0x4a = g0x4a = b0x4a = _mm_mullo_epi32(y0x4a, cyx4);
            r0x4a = _mm_add_epi32(r0x4a, _mm_mullo_epi32(vx4a, _mm_set1_epi32(crv)));
            r0x4a = _mm_add_epi32(r0x4a, rndx4);
            r0x4a = _mm_srai_epi32(r0x4a, in_sh);
            r0x4a = av_clip_int16_sse(r0x4a);

            r1x4a = g1x4a = b1x4a = _mm_mullo_epi32(y1x4a, cyx4);
            r1x4a = _mm_add_epi32(r1x4a, _mm_mullo_epi32(vx4a, _mm_set1_epi32(crv)));
            r1x4a = _mm_add_epi32(r1x4a, rndx4);
            r1x4a = _mm_srai_epi32(r1x4a, in_sh);
            r1x4a = av_clip_int16_sse(r1x4a);

            // g = av_clip_int16((y * cy + cgu * u + cgv * v + in_rnd) >> in_sh);
            g0x4a = _mm_add_epi32(g0x4a, _mm_mullo_epi32(ux4a, _mm_set1_epi32(cgu)));
            g0x4a = _mm_add_epi32(g0x4a, _mm_mullo_epi32(vx4a, _mm_set1_epi32(cgv)));
            g0x4a = _mm_add_epi32(g0x4a, rndx4);
            g0x4a = _mm_srai_epi32(g0x4a, in_sh);
            g0x4a = av_clip_int16_sse(g0x4a);

            g1x4a = _mm_add_epi32(g1x4a, _mm_mullo_epi32(ux4a, _mm_set1_epi32(cgu)));
            g1x4a = _mm_add_epi32(g1x4a, _mm_mullo_epi32(vx4a, _mm_set1_epi32(cgv)));
            g1x4a = _mm_add_epi32(g1x4a, rndx4);
            g1x4a = _mm_srai_epi32(g1x4a, in_sh);
            g1x4a = av_clip_int16_sse(g1x4a);

            // b = av_clip_int16((y * cy + cbu * u + in_rnd) >> in_sh);
            b0x4a = _mm_add_epi32(b0x4a, _mm_mullo_epi32(ux4a, _mm_set1_epi32(cbu)));
            b0x4a = _mm_add_epi32(b0x4a, rndx4);
            b0x4a = _mm_srai_epi32(b0x4a, in_sh);
            b0x4a = av_clip_int16_sse(b0x4a);

            b1x4a = _mm_add_epi32(b1x4a, _mm_mullo_epi32(ux4a, _mm_set1_epi32(cbu)));
            b1x4a = _mm_add_epi32(b1x4a, rndx4);
            b1x4a = _mm_srai_epi32(b1x4a, in_sh);
            b1x4a = av_clip_int16_sse(b1x4a);

            r0x4b = g0x4b = b0x4b = _mm_mullo_epi32(y0x4b, cyx4);
            r0x4b = _mm_add_epi32(r0x4b, _mm_mullo_epi32(vx4b, _mm_set1_epi32(crv)));
            r0x4b = _mm_add_epi32(r0x4b, rndx4);
            r0x4b = _mm_srai_epi32(r0x4b, in_sh);
            r0x4b = av_clip_int16_sse(r0x4b);

            r1x4b = g1x4b = b1x4b = _mm_mullo_epi32(y1x4b, cyx4);
            r1x4b = _mm_add_epi32(r1x4b, _mm_mullo_epi32(vx4b, _mm_set1_epi32(crv)));
            r1x4b = _mm_add_epi32(r1x4b, rndx4);
            r1x4b = _mm_srai_epi32(r1x4b, in_sh);
            r1x4b = av_clip_int16_sse(r1x4b);

            g0x4b = _mm_add_epi32(g0x4b, _mm_mullo_epi32(ux4b, _mm_set1_epi32(cgu)));
            g0x4b = _mm_add_epi32(g0x4b, _mm_mullo_epi32(vx4b, _mm_set1_epi32(cgv)));
            g0x4b = _mm_add_epi32(g0x4b, rndx4);
            g0x4b = _mm_srai_epi32(g0x4b, in_sh);
            g0x4b = av_clip_int16_sse(g0x4b);

            g1x4b = _mm_add_epi32(g1x4b, _mm_mullo_epi32(ux4b, _mm_set1_epi32(cgu)));
            g1x4b = _mm_add_epi32(g1x4b, _mm_mullo_epi32(vx4b, _mm_set1_epi32(cgv)));
            g1x4b = _mm_add_epi32(g1x4b, rndx4);
            g1x4b = _mm_srai_epi32(g1x4b, in_sh);
            g1x4b = av_clip_int16_sse(g1x4b);

            b0x4b = _mm_add_epi32(b0x4b, _mm_mullo_epi32(ux4b, _mm_set1_epi32(cbu)));
            b0x4b = _mm_add_epi32(b0x4b, rndx4);
            b0x4b = _mm_srai_epi32(b0x4b, in_sh);
            b0x4b = av_clip_int16_sse(b0x4b);

            b1x4b = _mm_add_epi32(b1x4b, _mm_mullo_epi32(ux4b, _mm_set1_epi32(cbu)));
            b1x4b = _mm_add_epi32(b1x4b, rndx4);
            b1x4b = _mm_srai_epi32(b1x4b, in_sh);
            b1x4b = av_clip_int16_sse(b1x4b);

            tonemap_int32x4_sse(r0x4a, g0x4a, b0x4a, r, g, b,
                                params->lin_lut, params->tonemap_lut, params->delin_lut,
                                params->coeffs, params->ocoeffs, params->desat, params->rgb2rgb_coeffs,
                                params->rgb2rgb_passthrough);
            tonemap_int32x4_sse(r1x4a, g1x4a, b1x4a, r1, g1, b1,
                                params->lin_lut, params->tonemap_lut, params->delin_lut,
                                params->coeffs, params->ocoeffs, params->desat, params->rgb2rgb_coeffs,
                                params->rgb2rgb_passthrough);
            tonemap_int32x4_sse(r0x4b, g0x4b, b0x4b, &r[4], &g[4], &b[4],
                                params->lin_lut, params->tonemap_lut, params->delin_lut,
                                params->coeffs, params->ocoeffs, params->desat, params->rgb2rgb_coeffs,
                                params->rgb2rgb_passthrough);
            tonemap_int32x4_sse(r1x4b, g1x4b, b1x4b, &r1[4], &g1[4], &b1[4],
                                params->lin_lut, params->tonemap_lut, params->delin_lut,
                                params->coeffs, params->ocoeffs, params->desat, params->rgb2rgb_coeffs,
                                params->rgb2rgb_passthrough);

            r0ox8 = _mm_lddqu_si128((const __m128i_u *)r);
            g0ox8 = _mm_lddqu_si128((const __m128i_u *)g);
            b0ox8 = _mm_lddqu_si128((const __m128i_u *)b);

            roax4 = _mm_cvtepi16_epi32(r0ox8);
            goax4 = _mm_cvtepi16_epi32(g0ox8);
            boax4 = _mm_cvtepi16_epi32(b0ox8);

            robx4 = _mm_unpackhi_epi16(r0ox8, zero128);
            gobx4 = _mm_unpackhi_epi16(g0ox8, zero128);
            bobx4 = _mm_unpackhi_epi16(b0ox8, zero128);

            yoax4 = _mm_mullo_epi32(roax4, _mm_set1_epi32(cry));
            yoax4 = _mm_add_epi32(yoax4, _mm_mullo_epi32(goax4, _mm_set1_epi32(cgy)));
            yoax4 = _mm_add_epi32(yoax4, _mm_mullo_epi32(boax4, _mm_set1_epi32(cby)));
            yoax4 = _mm_add_epi32(yoax4, _mm_set1_epi32(out_rnd));
            yoax4 = _mm_srai_epi32(yoax4, out_sh);
            yoax4 = _mm_add_epi32(yoax4, _mm_set1_epi32(params->out_yuv_off));

            yobx4 = _mm_mullo_epi32(robx4, _mm_set1_epi32(cry));
            yobx4 = _mm_add_epi32(yobx4, _mm_mullo_epi32(gobx4, _mm_set1_epi32(cgy)));
            yobx4 = _mm_add_epi32(yobx4, _mm_mullo_epi32(bobx4, _mm_set1_epi32(cby)));
            yobx4 = _mm_add_epi32(yobx4, _mm_set1_epi32(out_rnd));
            yobx4 = _mm_srai_epi32(yobx4, out_sh);
            yobx4 = _mm_add_epi32(yobx4, _mm_set1_epi32(params->out_yuv_off));

            y0ox8 = _mm_packus_epi32(yoax4, yobx4);
            y0ox8 = _mm_slli_epi16(y0ox8, out_sh2);
            _mm_storeu_si128((__m128i_u *) &dsty[x], y0ox8);

            r1ox8 = _mm_lddqu_si128((const __m128i_u *)r1);
            g1ox8 = _mm_lddqu_si128((const __m128i_u *)g1);
            b1ox8 = _mm_lddqu_si128((const __m128i_u *)b1);

            r1oax4 = _mm_cvtepi16_epi32(r1ox8);
            g1oax4 = _mm_cvtepi16_epi32(g1ox8);
            b1oax4 = _mm_cvtepi16_epi32(b1ox8);

            r1obx4 = _mm_unpackhi_epi16(r1ox8, zero128);
            g1obx4 = _mm_unpackhi_epi16(g1ox8, zero128);
            b1obx4 = _mm_unpackhi_epi16(b1ox8, zero128);

            y1oax4 = _mm_mullo_epi32(r1oax4, _mm_set1_epi32(cry));
            y1oax4 = _mm_add_epi32(y1oax4, _mm_mullo_epi32(g1oax4, _mm_set1_epi32(cgy)));
            y1oax4 = _mm_add_epi32(y1oax4, _mm_mullo_epi32(b1oax4, _mm_set1_epi32(cby)));
            y1oax4 = _mm_add_epi32(y1oax4, _mm_set1_epi32(out_rnd));
            y1oax4 = _mm_srai_epi32(y1oax4, out_sh);
            y1oax4 = _mm_add_epi32(y1oax4, _mm_set1_epi32(params->out_yuv_off));

            y1obx4 = _mm_mullo_epi32(r1obx4, _mm_set1_epi32(cry));
            y1obx4 = _mm_add_epi32(y1obx4, _mm_mullo_epi32(g1obx4, _mm_set1_epi32(cgy)));
            y1obx4 = _mm_add_epi32(y1obx4, _mm_mullo_epi32(b1obx4, _mm_set1_epi32(cby)));
            y1obx4 = _mm_add_epi32(y1obx4, _mm_set1_epi32(out_rnd));
            y1obx4 = _mm_srai_epi32(y1obx4, out_sh);
            y1obx4 = _mm_add_epi32(y1obx4, _mm_set1_epi32(params->out_yuv_off));

            y1ox8 = _mm_packus_epi32(y1oax4, y1obx4);
            y1ox8 = _mm_slli_epi16(y1ox8, out_sh2);
            _mm_storeu_si128((__m128i_u *) &dsty[x + dstlinesize[0] / 2], y1ox8);

            ravgx4 = _mm_hadd_epi32(roax4, robx4);
            ravgx4 = _mm_add_epi32(ravgx4, _mm_hadd_epi32(r1oax4, r1obx4));
            ravgx4 = _mm_add_epi32(ravgx4, _mm_set1_epi32(2));
            ravgx4 = _mm_srai_epi32(ravgx4, 2);

            gavgx4 = _mm_hadd_epi32(goax4, gobx4);
            gavgx4 = _mm_add_epi32(gavgx4, _mm_hadd_epi32(g1oax4, g1obx4));
            gavgx4 = _mm_add_epi32(gavgx4, _mm_set1_epi32(2));
            gavgx4 = _mm_srai_epi32(gavgx4, 2);

            bavgx4 = _mm_hadd_epi32(boax4, bobx4);
            bavgx4 = _mm_add_epi32(bavgx4, _mm_hadd_epi32(b1oax4, b1obx4));
            bavgx4 = _mm_add_epi32(bavgx4, _mm_set1_epi32(2));
            bavgx4 = _mm_srai_epi32(bavgx4, 2);

            uoax4 = _mm_add_epi32(_mm_set1_epi32(out_rnd), _mm_mullo_epi32(ravgx4, _mm_set1_epi32(cru)));
            uoax4 = _mm_add_epi32(uoax4, _mm_mullo_epi32(gavgx4, _mm_set1_epi32(ocgu)));
            uoax4 = _mm_add_epi32(uoax4, _mm_mullo_epi32(bavgx4, _mm_set1_epi32(cburv)));
            uoax4 = _mm_srai_epi32(uoax4, out_sh);
            uoax4 = _mm_add_epi32(uoax4, _mm_set1_epi32(out_uv_offset));

            voax4 = _mm_add_epi32(_mm_set1_epi32(out_rnd), _mm_mullo_epi32(ravgx4, _mm_set1_epi32(cburv)));
            voax4 = _mm_add_epi32(voax4, _mm_mullo_epi32(gavgx4, _mm_set1_epi32(ocgv)));
            voax4 = _mm_add_epi32(voax4, _mm_mullo_epi32(bavgx4, _mm_set1_epi32(cbv)));
            voax4 = _mm_srai_epi32(voax4, out_sh);
            voax4 = _mm_add_epi32(voax4, _mm_set1_epi32(out_uv_offset));

            uvoax4 = _mm_unpacklo_epi32(uoax4, voax4);
            uvobx4 = _mm_unpackhi_epi32(uoax4, voax4);
            uvox8 = _mm_packus_epi32(uvoax4, uvobx4);
            uvox8 = _mm_slli_epi16(uvox8, out_sh2);
            _mm_storeu_si128((__m128i_u *) &dstuv[x], uvox8);
        }
    }

    // Process remaining pixels cannot fill the full simd register with scalar version
    if (remainw) {
        int offset = width & (int)0xfffffff8;
        rdsty += offset;
        rdstuv += offset;
        rsrcy += offset;
        rsrcuv += offset;
        tonemap_frame_p016_p010_2_p016_p010(rdsty, rdstuv,
                                            rsrcy, rsrcuv,
                                            dstlinesize, srclinesize,
                                            dstdepth, srcdepth,
                                            remainw, rheight, params);
    }
#endif // ENABLE_TONEMAPX_SSE_INTRINSICS
}
